/**

    Copyright 2007, 2010 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.metadata;

import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Class <code>SchemaDefinition</code> models the schema
 * definition. Istances are created with a factory method.
 *
 * @author <a href="mailto:gianfranco.boccalon@eng.it">Gianfranco
 * Boccalon</a> original code
 * @author <a href="mailto:saint@eng.it">Gian Uberto Lauri</a> port of
 * the factory method to W3C API.
 * @version $Revision$
 */
public class SchemaDefinition {
	private String name;
	private List<Field> fields;
	private List<Column> columns;

	/**
	 * <code>getFields</code> write access to the fields attribute
	 *
	 * @return a <code>List<Field></code> the fields values
	 */
	public List<Field> getFields() {
		return fields;
	}

	/**
	 * <code>setFields</code> write access to the fields attribute
	 *
	 * @param fields a <code>List<Field></code> with the new fields
	 * values
	 */
	public void setFields(List<Field> fields) {
		this.fields = fields;
	}

	/**
	 * <code>getColumns</code> write access to the columns attribute
	 *
	 * @return a <code>List<Column></code> the columns values
	 */
	public List<Column> getColumns() {
		return columns;
	}

	/**
	 * <code>setColumns</code> write access to the columns attribute
	 *
	 * @param fields a <code>List<Column></code> with the new columns
	 * values
	 */
	public void setColumns(List<Column> columns) {
		this.columns = columns;
	}

	/**
	 * <code>parseSchema</code> Factory method for the class
	 *
	 * @param xml a <code>String</code> 
	 * @return a <code>SchemaDefinition</code> 
	 * @exception ParserConfigurationException 
	 * @exception DOMException 
	 * @exception IOException 
	 * @exception SAXException 
	 */
	public static SchemaDefinition parseSchema(String xml) throws ParserConfigurationException, DOMException, IOException, SAXException {
		// Parse the XML as a W3C document.
		DocumentBuilderFactory xmlFact = DocumentBuilderFactory.newInstance();
		xmlFact.setNamespaceAware(true);
		DocumentBuilder builder = xmlFact.newDocumentBuilder();
		InputSource source = new InputSource(new StringReader(xml));
        Document document = builder.parse(source);

		SchemaDefinition formDescription = null;
		if (document != null) {
			formDescription = new SchemaDefinition();

			NodeList fieldsXml = document.getFirstChild().getChildNodes();

			List<Field> fields = null;
			List<Column> columns = null;

			int maxLen = fieldsXml.getLength();
			for (int i=0; i < maxLen; i++) {
				BaseField baseField = null;
				Node fieldElement = fieldsXml.item(i);
				if( fieldElement instanceof Element) {
					if (fieldElement.getNodeName().equals("field")) {
						baseField = new Field();
						if (fields == null)
							fields = new ArrayList<Field>();
					} else {
						baseField = new Column();
						((Column) baseField).setEditable(
								// Boolean.parseBoolean(.getAttributes("editable"));
								Boolean.parseBoolean(getAttribute(fieldElement,"editable")));
						if (columns == null)
							columns = new ArrayList<Column>();
					}
					baseField.setDefaultValue(getAttribute(fieldElement, "defaultValue"));
					baseField.setCombo(Boolean.parseBoolean(getAttribute(fieldElement,"combo")));
					baseField.setMandatory(Boolean.parseBoolean(getAttribute(fieldElement,"mandatory")));
					baseField.setType(getAttribute(fieldElement,"type"));

					String validator = getAttribute(fieldElement,"validator");
					// Reg-ex associations
					if ("None".equals(validator))
						baseField.setValidator(".*");

					else if ("Number".equals(validator))
						baseField.setValidator("[\\d\\.\\,]*");

					else if ("Text".equals(validator))
						baseField.setValidator("[a-zA-Z \\.\\,\\:\\;]*");

					else if ("Number & Text".equals(validator))
						baseField.setValidator("\\w*");

					else if ("Telephone Number".equals(validator))
						baseField.setValidator("\\+{0,1}\\d{10,14}");

					else if ("E-mail".equals(validator))
						baseField.setValidator("[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\\.[A-Za-z]{2,4}");

					baseField.setName(getAttribute(fieldElement,"name"));
					baseField.setId(getAttribute(fieldElement,"id"));

					String tmp = getAttribute(fieldElement,"length");
					if (tmp != null)
						baseField.setLength(Integer.parseInt(tmp));

					tmp = getAttribute(fieldElement,"precision");
					if (tmp != null)
						baseField.setPrecision(Integer.parseInt(tmp));

					if (baseField.isCombo()) {
						Iterator<Node> comboElements = elementIterator(fieldElement, "item");
						List<ComboItem> items = null;
						if (comboElements.hasNext()) {
							items = new ArrayList<ComboItem>();
						}
						while (comboElements.hasNext()) {
							Node itemElement = comboElements.next();
							ComboItem item = new ComboItem();
							item.setLabel(getAttribute(itemElement,"name"));
							item.setValue(getAttribute(itemElement,"value"));
							items.add(item);
						}
						baseField.setItems(items);
					}
					if (fieldElement.getNodeName().equals("field")) {
						fields.add((Field) baseField);
					} else {
						columns.add((Column) baseField);
					}
				}
			}
			formDescription.setFields(fields);
			formDescription.setColumns(columns);
		}
		return formDescription;
	}

	/*
	 * Here we have to replace the node method with
	 * a periphrasis. We could have done this in a more OO way, but
	 * this keeps things clearer with the various classloaders...
	 */
	private static String getAttribute(Node node, String name) 
		throws DOMException {
		return node.getAttributes().getNamedItem(name)
			.getNodeValue();
	}

	private static Iterator<Node> elementIterator(Node node, String name) 
		throws DOMException {
		NodeList nl = node.getChildNodes();
		int maxLen = nl.getLength();
		ArrayList<Node> rvList = new ArrayList<Node>(maxLen);

		for (int i = 0; i < maxLen; i++) {
			Node n = nl.item(i);
			if(name.equals(n.getNodeName()))
				rvList.add(n);
		}

		return rvList.iterator();
	}

	/**
	 * <code>getName</code> read access to the name attribute
	 *
	 * @return a <code>String</code> the value of the name attribute
	 */
	public String getName() {
		return name;
	}

	/**
	 * <code>setName</code> write access to the name attribute
	 *
	 * @param name a <code>String</code> the new value for the name attribute
	 */
	public void setName(String name) {
		this.name = name;
	}
}
