/**

    Copyright 2007, 2009 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
 **/
package org.spagic.workflow.api.jbpm;

import java.io.File;
import java.lang.reflect.Constructor;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.xml.namespace.QName;



import org.hibernate.Session;
import org.hibernate.Transaction;
import org.jbpm.JbpmConfiguration;
import org.jbpm.JbpmContext;
import org.jbpm.graph.def.ProcessDefinition;
import org.jbpm.graph.exe.Token;
import org.jbpm.job.executor.JobExecutor;
import org.jbpm.taskmgmt.exe.PooledActor;
import org.jbpm.taskmgmt.exe.TaskInstance;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic.metadb.base.ProcessExecutionStateConstants;
import org.spagic.metadb.base.ProcessInstanceStateConstants;
import org.spagic.metadb.dbutils.AuditDBUtils;
import org.spagic.metadb.dbutils.HibernateUtil;
import org.spagic.metadb.model.Process;
import org.spagic.metadb.model.ProcessExecutionState;
import org.spagic.metadb.model.ProcessInstance;
import org.spagic.metadb.model.Property;
import org.spagic.metadb.model.Step;
import org.spagic.workflow.api.IControlAPI;
import org.spagic.workflow.api.IControlExtendedAPI;
import org.spagic.workflow.api.IProcessEngine;
import org.spagic.workflow.api.IProfileService;
import org.spagic.workflow.api.IQueryAPI;
import org.spagic.workflow.api.SpagicWorkflowException;
import org.spagic.workflow.api.Task;
import org.spagic.workflow.api.Variable;
import org.spagic.workflow.logging.log.FaultLog;

import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.io.xml.Dom4JDriver;


// TODO: Auto-generated Javadoc
/**
 * The Class ProcessEngine.
 */
public class ProcessEngine implements IProcessEngine {

	/** The Constant log. */
	private static Logger log = LoggerFactory.getLogger(ProcessEngine.class);

	/** The Constant JBPM_PREFIX. */
	public static final String JBPM_PREFIX = "JBPM:".intern();

	/** The Constant PREFIX_NAMESPACE. */
	public static final String PREFIX_NAMESPACE = "urn:eng:spagic:processes:".intern();

	/** The Constant VERSION_LABEL. */
	public static final String VERSION_LABEL = "_v_".intern();

	/** The Constant PROPERTY_VARIABLE_PREFIX. */
	public static final String PROPERTY_VARIABLE_PREFIX = "Var[".intern();

	/** The Constant PROPERTY_VARIABLE_PREFIX. */
	public static final String TASK_PROPERTY_VARIABLE_PREFIX = "_[Task]_".intern();

	/** The Constant MESSAGE_ID_PREFIX_TOKEN. */
	public static final String MESSAGE_ID_PREFIX_TOKEN = "TOK:".intern();

	/** The Constant MESSAGE_ID_PREFIX_START_NODE. */
	public static final String MESSAGE_ID_PREFIX_START_NODE = ":START_NODE:".intern();

	/** The Constant MESSAGE_ID_PREFIX_DESTINATION_NODE. */
	public static final String MESSAGE_ID_PREFIX_DESTINATION_NODE = ":DEST_NODE:".intern();

	/** The Constant MESSAGE_ID_PREFIX_TIMESTAMP. */
	public static final String MESSAGE_ID_PREFIX_TIMESTAMP = ":".intern();
	
	/** The variable name for storing the process instance id */
	public static final String PROCESS_INSTANCE_ID_VARIABLE = "PROCESS_INSTANCE_ID_VARIABLE".intern();
	
	/** Constants for managing completion percentage within task instances */
	public static final String COMMENT_SEPARATION = "#";
	public static final String COMPLETION_PERCENTAGE_COMMENT = "COMPLETION_PERCENTAGE";

	/** The Constant sdf. */
	public static final SimpleDateFormat sdf = new SimpleDateFormat("dd/MM/yyyy HH:mm:ss");

	/** The Constant sdfLight. */
	public static final SimpleDateFormat sdfLight = new SimpleDateFormat("dd/MM/yyyy");

	/** The control api. */
	private IControlAPI controlAPI = null;

	/** The extended control api. */
	private IControlExtendedAPI controlExtendedAPI = null;

	/** The query api. */
	private IQueryAPI queryAPI = null;
		
	private IProfileService profileService = null;

	/** The jbpm configuration. */
	protected static JbpmConfiguration jbpmConfiguration = null;

	/** The process execution states. */
	private static Map<Long, ProcessExecutionState> processExecutionStates = null;

	
	/** Locking utilities on root tokens for multithreaded usage */
	private static ConcurrentHashMap<Long, Object> lockTokens = new ConcurrentHashMap();
	
	private static void lockToken(Long tokenId){	
		if (!lockTokens.containsKey(tokenId)){
			synchronized (lockTokens) {
				lockTokens.put(tokenId, new Object());
				log.info("["+Thread.currentThread().getName()+"] Token ["+tokenId+"] is NOT LOCKED -> LOCKED ACQUIRED");
			}
		}else{
			log.info("["+Thread.currentThread().getName()+"] Token ["+tokenId+"] is LOCKED -> WAITING");
			Object lockToken = lockTokens.get(tokenId);
			synchronized (lockToken) {
				try{
					lockToken.wait();
				}catch (InterruptedException e) {
					// TODO: handle exception
				}
			}
		}
		
	}
	
	//
	// Utility Method to handle resources
	// 
	private static void closeJbpmContextAndUnlockToken(JbpmContext context, Long tokenid){
		try {
			if (context != null)
				context.close();
		} catch (Throwable th) {
			log.error("Error during resources finalization", th);
		}
		
		if (tokenid != null){
			Object o = lockTokens.remove(tokenid);
			if (o != null){
				synchronized (o) {
					log.info("["+Thread.currentThread().getName()+"] Closing Context and unlocking token ["+tokenid+"]");
					o.notify();
				}
			}
		}
	}
	
	private static void closeSession(Session aSession){
		if (aSession != null && aSession.isOpen()) {
			aSession.close();
		}
	}
	
	
	private static void finalizeResources(JbpmContext jbpmContext, Session spagicSession, Transaction spagicTx, Long rootTokenId, boolean faulted, Throwable error){
		try {
			if (faulted){
				// Fault
				if (jbpmContext != null)
					jbpmContext.setRollbackOnly();
				if ( (spagicTx != null ) && spagicTx.isActive())
					spagicTx.rollback();
			}else{
				// Success
				if ( (spagicTx != null ) && spagicTx.isActive())
					spagicTx.commit();
			}
		} catch (Throwable th) {
			log.error("Error during resources finalization", th);
		}
		
		try {
			closeJbpmContextAndUnlockToken(jbpmContext, rootTokenId);
		} catch (Throwable th) {
			log.error("Error during resources finalization", th);
		}
		
		try {
			closeSession(spagicSession);
		} catch (Throwable th) {
			log.error("Error during resources finalization", th);
		}
		
		if (faulted && (error != null)){
			throw new SpagicWorkflowException(error.getMessage(), error);
		}
	}
	
	//
	// Handling Spagic Hibernate Session and JBPM Context
	//
	/**
	 * Gets the session.
	 * 
	 * @return the session
	 */
	protected static Session getSession() {
		Session aSession = HibernateUtil.getSessionFactory().openSession();
		return aSession;
	}
	
	
	private static JbpmContext getJbpmContext(){
		JbpmContext jbpmContext = jbpmConfiguration.createJbpmContext();
		return jbpmContext;	
	}
	
	public static Object doWithinJBPMAndSpagic(String invokingMethodName, ProcessEngine processEngine, ISpagicJBPMCommand bpmEngineCommand, ProcessInstanceExecutionContext iec, boolean lockRootToken){
		AtomicBoolean faulted = new AtomicBoolean();
		return doWithinJBPMAndSpagic(invokingMethodName, processEngine, bpmEngineCommand, iec, lockRootToken, faulted);
	}
	
	public static Object doWithinJBPMAndSpagic(String invokingMethodName, ProcessEngine processEngine, ISpagicJBPMCommand bpmEngineCommand, ProcessInstanceExecutionContext iec, boolean lockRootToken, AtomicBoolean faulted){

		JbpmContext jbpmContext = null;
		Session spagicSession = null;
		Transaction spagicTx = null;
		Object result = null;
		Throwable error = null;

		Long jbpmRootTokenId = (iec != null) ? iec.getJbpmRootTokenId() : null;
		faulted.set(false);
		try {
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Start");
			if (lockRootToken){
				log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Locking Token ["+iec.getJbpmRootTokenId()+"]");
				if (jbpmRootTokenId != null)
					ProcessEngine.lockToken(jbpmRootTokenId);
				else
					log.error("Lock on tocken requested but cannot retrieve Process Instance Execution Context");
			}
			
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Getting JBPM Context");
			jbpmContext = getJbpmContext();
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Getting Spagic Session");
			spagicSession = getSession();
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Begin Transaction");
			spagicTx = spagicSession.beginTransaction();
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Executing Command ");
			result = bpmEngineCommand.execute(processEngine, jbpmContext,spagicSession, iec );
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Command Execution OK ");

		}catch (Throwable t) {
			
			log.error("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Execution Faulted ", t);
			faulted.set(true);
			error = t;
			
		}finally{
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Finalize Resources ");
			ProcessEngine.finalizeResources(jbpmContext, spagicSession, spagicTx, jbpmRootTokenId, faulted.get(), error);
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Returning Result ");
		}
		return result;
		
	}
	
	
	public static Object doQueryWithinJBPMAndSpagic(String invokingMethodName, ProcessEngine processEngine, ISpagicJBPMCommand bpmEngineCommand, ProcessInstanceExecutionContext iec){

		JbpmContext jbpmContext = null;
		Session spagicSession = null;

		Object result = null;
		boolean faulted = false;
		Throwable error = null;

		try {
			
			
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Getting JBPM Context");
			jbpmContext = getJbpmContext();
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Getting Spagic Session");
			spagicSession = getSession();
			
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Executing Command ");
			result = bpmEngineCommand.execute(processEngine, jbpmContext,spagicSession, iec );
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Command Execution OK ");

		}catch (Throwable t) {
			log.error("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Execution Faulted ", t);
			faulted = true;
			error = t;
			
		}finally{
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Finalize Resources ");
			ProcessEngine.finalizeResources(jbpmContext, spagicSession, null, null, faulted, error);
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Returning Result ");
		}
		return result;
		
	}
	
	public static Object doQueryWithinSpagic(String invokingMethodName, ProcessEngine processEngine, ISpagicJBPMCommand bpmEngineCommand, ProcessInstanceExecutionContext iec){

		Session spagicSession = null;

		Object result = null;
		boolean faulted = false;
		Throwable error = null;

		try {
			log.info("["+invokingMethodName+"] -> doQueryWithinSpagic - Getting Spagic Session");
			spagicSession = getSession();
			
			log.info("["+invokingMethodName+"] -> doQueryWithinSpagic - Executing Command ");
			result = bpmEngineCommand.execute(processEngine, null,spagicSession, iec );
			log.info("["+invokingMethodName+"] -> doQueryWithinSpagic - Command Execution OK ");

		}catch (Throwable t) {
			log.error("["+invokingMethodName+"] -> doQueryWithinSpagic - Execution Faulted ", t);
			faulted = true;
			error = t;
			
		}finally{
			log.info("["+invokingMethodName+"] -> doQueryWithinSpagic - Finalize Resources ");
			ProcessEngine.finalizeResources(null, spagicSession, null, null, faulted, error);
			log.info("["+invokingMethodName+"] -> doQueryWithinSpagic - Returning Result ");
		}
		return result;
	}
	
	public static Object doWithinJBPM(String invokingMethodName, ProcessEngine processEngine, ISpagicJBPMCommand bpmEngineCommand, ProcessInstanceExecutionContext iec, boolean lockRootToken){
		AtomicBoolean faulted = new AtomicBoolean();
		return doWithinJBPM(invokingMethodName, processEngine, bpmEngineCommand, iec, lockRootToken, faulted);
	}
		
	public static Object doWithinJBPM(String invokingMethodName, ProcessEngine processEngine, ISpagicJBPMCommand bpmEngineCommand, ProcessInstanceExecutionContext iec, boolean lockRootToken, AtomicBoolean faulted){

		JbpmContext jbpmContext = null;
		Object result = null;
		Throwable error = null;

		Long jbpmRootTokenId = iec != null ? iec.getJbpmRootTokenId() : null;
		faulted.set(false);
		try {
			log.info("["+invokingMethodName+"] -> doWithinJBPM - Start");
			log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Start");
			if (lockRootToken){
				log.info("["+invokingMethodName+"] -> doWithinJBPMAndSpagic - Locking Token ["+iec.getJbpmRootTokenId()+"]");
				if (jbpmRootTokenId != null)
					ProcessEngine.lockToken(jbpmRootTokenId);
				else
					log.error("Lock on tocken requested but cannot retrieve Process Instance Execution Context");
			}
			
			log.info("["+invokingMethodName+"] -> doWithinJBPM - Getting JBPM Context");
			jbpmContext = getJbpmContext();
			log.info("["+invokingMethodName+"] -> doWithinJBPM - Executing Command ");
			result = bpmEngineCommand.execute(processEngine, jbpmContext, null, iec );
			log.info("["+invokingMethodName+"] -> doWithinJBPM - Command Execution OK ");

		}catch (Throwable t) {
			log.error("["+invokingMethodName+"] -> doWithinJBPM - Execution Faulted ", t);
			faulted.set(true);
			error = t;
			
		}finally{
			log.info("["+invokingMethodName+"] -> doWithinJBPM - Finalize Resources ");
			ProcessEngine.finalizeResources(jbpmContext, null, null, jbpmRootTokenId, faulted.get(), error);
			log.info("["+invokingMethodName+"] -> doWithinJBPM - Returning Result ");
		}
		return result;
	}

	public static Object doWithinSpagic(String invokingMethodName, ProcessEngine processEngine, ISpagicJBPMCommand bpmEngineCommand, ProcessInstanceExecutionContext iec){
		AtomicBoolean faulted = new AtomicBoolean();
		return doWithinSpagic(invokingMethodName, processEngine, bpmEngineCommand, iec, faulted);		
	}
		
	public static Object doWithinSpagic(String invokingMethodName, ProcessEngine processEngine, ISpagicJBPMCommand bpmEngineCommand, ProcessInstanceExecutionContext iec, AtomicBoolean faulted){

		Session spagicSession = null;
		Transaction spagicTx = null;
		Object result = null;
		Throwable error = null;

		faulted.set(false);
		try {
			log.info("["+invokingMethodName+"] -> doWithinSpagic - Getting Spagic Session");
			spagicSession = getSession();
			log.info("["+invokingMethodName+"] -> doWithinSpagic - Begin Transaction");
			spagicTx = spagicSession.beginTransaction();
			log.info("["+invokingMethodName+"] -> doWithinSpagic - Executing Command ");
			
			result = bpmEngineCommand.execute(processEngine, null, spagicSession, iec );
			log.info("["+invokingMethodName+"] -> doWithinSpagic - Command Execution OK ");

		}catch (Throwable t) {
			
			log.error("["+invokingMethodName+"] -> doWithinSpagic - Execution Faulted ", t);
			faulted.set(true);
			error = t;
			
		}finally {
			
			log.info("["+invokingMethodName+"] -> doWithinSpagic - Finalize Resources ");
			ProcessEngine.finalizeResources(null, spagicSession, spagicTx, null, faulted.get(), error);
			log.info("["+invokingMethodName+"] -> doWithinSpagic - Returning Result ");
			
		}
		return result;
		
	}
	
	
	
	
	/**
	 * Instantiates a new process engine.
	 */
	public ProcessEngine() {
		this(true);
	}
	
	public ProcessEngine(boolean startJobExecutor) {
		super();
		jbpmConfiguration = JbpmConfiguration.getInstance();
		if (startJobExecutor) {
			jbpmConfiguration.startJobExecutor();
		}
		
		controlAPI = new ControlAPI(jbpmConfiguration, this);
		controlExtendedAPI = new ControlExtendedAPI(jbpmConfiguration, this);
		queryAPI = new QueryAPI(jbpmConfiguration, this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IProcessEngine#getControlAPI()
	 */
	public IControlAPI getControlAPI() {
		return controlAPI;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IProcessEngine#getControlExtendedAPI()
	 */
	public IControlExtendedAPI getControlExtendedAPI() {
		return controlExtendedAPI;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IProcessEngine#getQueryAPI()
	 */
	public IQueryAPI getQueryAPI() {
		return queryAPI;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IProcessEngine#stop()
	 */
	public void stop() {
		JobExecutor je = jbpmConfiguration.getJobExecutor();
		if ((je != null) && je.isStarted()) {
			je.stop();
		}
	}

	/**
	 * Gets the name space for process.
	 * 
	 * @param processName
	 *          the process name
	 * @param currentVersion
	 *          the current version
	 * 
	 * @return the name space for process
	 */
	public static String getNameSpaceForProcess(String processName, String currentVersion) {
		return PREFIX_NAMESPACE + processName + ":v" + currentVersion;
	}

	/**
	 * Gets the process q name.
	 * 
	 * @param processName
	 *          the process name
	 * @param currentVersion
	 *          the current version
	 * 
	 * @return the process q name
	 */
	public static QName getProcessQName(String processName, String currentVersion) {
		QName qName = new QName(getNameSpaceForProcess(processName, currentVersion), processName);
		return qName;
	}

	/**
	 * Extract jbpm process instance id from id correlation.
	 * 
	 * @param idCorrelation
	 *          the id correlation
	 * 
	 * @return the long
	 */
	public static Long extractJbpmProcessInstanceIdFromIdCorrelation(String idCorrelation) {
		String jbpmId = idCorrelation.substring(JBPM_PREFIX.length());
		return Long.valueOf(jbpmId);
	}

	/**
	 * Gets the jbpm process instance id.
	 * 
	 * @param spagicProcessInstanceID
	 *          the spagic process instance id
	 * 
	 * @return the jbpm process instance id */

	
	public static ProcessInstanceExecutionContext buildProcessInstanceExecutionContext(final long spagicProcessInstanceId){
		
		return (ProcessInstanceExecutionContext)
				ProcessEngine.doWithinJBPMAndSpagic("buildProcessInstanceExecutionContext("+spagicProcessInstanceId+")", 
				null, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						ProcessInstance spagicProcessInstance = AuditDBUtils.getProcessInstanceById(spagicSession, spagicProcessInstanceId);
						long jbpmProcessInstanceId = extractJbpmProcessInstanceIdFromIdCorrelation(spagicProcessInstance.getIdCorrelation());
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstance(jbpmProcessInstanceId); 
						long jbpmRootTokenId = jbpmProcessInstance.getRootToken().getId();
						
						ProcessInstanceExecutionContext piec = new ProcessInstanceExecutionContext(spagicProcessInstanceId, jbpmProcessInstanceId, jbpmRootTokenId);
						return piec;
					}
				}, 
				null, false);
		
	}
	

	/**
	 * Gets the variable name from property code.
	 * 
	 * @param propertyCode
	 *          the property code
	 * 
	 * @return the variable name from property code
	 */
	public static String getVariableNameFromPropertyCode(String propertyCode) {
		if (propertyCode.startsWith(TASK_PROPERTY_VARIABLE_PREFIX))
			return propertyCode;
		String variableName = propertyCode.substring(PROPERTY_VARIABLE_PREFIX.length(), propertyCode.length() - 1);
		return variableName;
	}

	/**
	 * Gets the variable value from property value.
	 * 
	 * @param propertyValue
	 *          the property value
	 * 
	 * @return the variable value from property value
	 */
	public static Object getVariableValueFromPropertyValue(String propertyValue) {

		String className = null;
		String stringValue = null;
		if (propertyValue.indexOf("(") < 0) {
			className = propertyValue;
			stringValue = "";
		} else {
			className = propertyValue.substring(0, propertyValue.indexOf("("));
			stringValue = propertyValue.substring(propertyValue.indexOf("(") + 1, propertyValue.indexOf(")"));
		}

		if (className.equalsIgnoreCase("java.util.Date")) {

			Date date = new Date();
			try {
				date = sdf.parse(stringValue);
			} catch (Exception e) {
				try {
					date = sdfLight.parse(stringValue);
				} catch (Exception xe) {
					xe.printStackTrace();
				}
			}
			return date;
		} else {
			Object result = null;
			try {
				Class objClass = Class.forName(className);
				Constructor objConstructor = objClass.getConstructor((Class[]) null);
				result = objConstructor.newInstance(null);
			} catch (Exception e) {
				e.printStackTrace();
			}
			return result;
		}

	}

	// --- METHODS NOT VISIBLE IN THE INTERFACES

	/**
	 * Creates the spagic process.
	 * 
	 * @param jBPMProcessInstance
	 *          the j bpm process instance
	 * @param jBPMParentProcessInstance
	 *          the j bpm parent process instance
	 * @param aSession
	 *          the a session
	 * 
	 * @return the org.spagic.metadb.model. process instance
	 */
	public static org.spagic.metadb.model.ProcessInstance createSpagicProcess(
			org.jbpm.graph.exe.ProcessInstance jBPMProcessInstance,
			org.jbpm.graph.exe.ProcessInstance jBPMParentProcessInstance, Session aSession) {
		try {
			String correlationID = ProcessEngine.createSpagicCorrelationId(jBPMProcessInstance);

			// Retrieves the process name
			ProcessDefinition processDefinition = null;
			if (jBPMProcessInstance != null)
				processDefinition = jBPMProcessInstance.getProcessDefinition();
			String processName = null;
			if (processDefinition != null)
				processName = processDefinition.getName();

			// Only the latest version can be instantiated
			Process process = ProcessEngine.getLastDefinitionForProcess(aSession, processName);

			// Retrieve parent process, if any
			org.spagic.metadb.model.ProcessInstance parentProcessInstance = null;
			if (jBPMParentProcessInstance != null) {
				parentProcessInstance = ProcessEngine.getSpagicProcessInstanceFromJBPMProcessInstance(
						jBPMParentProcessInstance, aSession);
			}

			// Creation of a new Spagic process instance
			org.spagic.metadb.model.ProcessInstance processInstance = ProcessEngine.createProcessInstance(process,
					correlationID, parentProcessInstance, aSession);

			// /////////////////////////// TEMP: for variables management, for lazy
			// loading
			org.spagic.metadb.model.Process spagicProcess = null;
			if (processInstance != null)
				spagicProcess = processInstance.getProcess();
			Set<org.spagic.metadb.model.PropertyProcess> properties = null;
			if (spagicProcess != null)
				properties = spagicProcess.getPropertyProcesses();
			if (properties != null)
				for (org.spagic.metadb.model.PropertyProcess property : properties) {
					org.spagic.metadb.model.Property prop = property.getProperty();
					prop.getCode();
				}
			// ///////////////////////////

			return processInstance;
		} catch (Exception e) {
			log.error("createSpagicProcess ", e);
			return null;
		} finally {

		}
	}

	/**
	 * Create process instance global variables.
	 * 
	 * @param spagicProcessInstance
	 *          the spagic process instance
	 * @param jbpmProcessInstance
	 *          the jbpm process instance
	 */
	public static void createSpagicVariables(org.spagic.metadb.model.ProcessInstance spagicProcessInstance,
			org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance) {
		// Create global variables
		List<org.spagic.metadb.model.Property> spagicVariablesProperties = getSpagicVariablesProperties(spagicProcessInstance);
		for (org.spagic.metadb.model.Property property : spagicVariablesProperties) {
			String propName = property.getCode();
			String varName = ProcessEngine.getVariableNameFromPropertyCode(propName);
			jbpmProcessInstance.getContextInstance().setVariable(varName,
					ProcessEngine.getVariableValueFromPropertyValue(property.getValue()));
		}
		org.spagic.metadb.model.Process spagicProcess = spagicProcessInstance.getProcess();
		for (Step step : spagicProcess.getSteps()) {
			String taskname = step.getName();
			taskname = taskname.substring(taskname.indexOf('.') + 1);
			taskname = taskname.substring(0, taskname.indexOf("_v_"));
			for (org.spagic.metadb.model.PropertyStep property : step.getPropertySteps()) {
				Property propertyTask = property.getProperty();
				String propName = propertyTask.getCode();
				// TaskVar[pippoVarTask]
				if (propName.startsWith("TaskVar[")) {
					propName = propName.substring(8, propName.length() - 1);
					jbpmProcessInstance.getContextInstance().setVariable("_[Task]_" + taskname + "." + propName,
							getVariableValueFromPropertyValue(propertyTask.getValue()));
				}
			}
		}
		
		// Create also a global variable containing the Spagic process instance id
		jbpmProcessInstance.getContextInstance().setVariable(PROCESS_INSTANCE_ID_VARIABLE, spagicProcessInstance.getIdProcessInstance());
	}

	/**
	 * Gets the spagic variables properties.
	 * 
	 * @param spagicProcessInstance
	 *          the spagic process instance
	 * 
	 * @return the spagic variables properties
	 */
	public static List<org.spagic.metadb.model.Property> getSpagicVariablesProperties(
			org.spagic.metadb.model.ProcessInstance spagicProcessInstance) {
		org.spagic.metadb.model.Process spagicProcess = spagicProcessInstance.getProcess();
		List<org.spagic.metadb.model.Property> result = new ArrayList<org.spagic.metadb.model.Property>();

		Set<org.spagic.metadb.model.PropertyProcess> properties = spagicProcess.getPropertyProcesses();
		for (org.spagic.metadb.model.PropertyProcess property : properties) {
			if (property.getProperty().getCode().startsWith(PROPERTY_VARIABLE_PREFIX)) {
				result.add(property.getProperty());
			}
		}
		return result;
	}

	/**
	 * Find task instance with name.
	 * 
	 * @param jbpmProcessInstance
	 *          the jbpm process instance
	 * @param taskName
	 *          the task name
	 * 
	 * @return the task instance
	 */
	public static TaskInstance findTaskInstanceWithName(org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance, String taskName) {
		Collection<TaskInstance> taskInstances = (Collection<TaskInstance>) jbpmProcessInstance.getTaskMgmtInstance()
				.getTaskInstances();
		for (TaskInstance ti : taskInstances) {
			if (ti.getName().equals(taskName) && !ti.hasEnded()) {
				return ti;
			}
		}
		return null;
	}

	/**
	 * Gets the process instance status variable name.
	 * 
	 * @param processInstanceId
	 *          the process instance id
	 * 
	 * @return the process instance status variable name
	 */
	public static String getProcessInstanceStatusVariableName(long processInstanceId) {
		Session aSession = null;
		try {
			aSession = getSession();
			ProcessInstance processInstance = AuditDBUtils.getProcessInstanceById(aSession, processInstanceId);

			Property processInstanceStatusVariableNameAsProperty = AuditDBUtils.getProcessProperty(aSession, processInstance
					.getProcess(), "ProcessInstanceStatusVariableName");
			if (processInstanceStatusVariableNameAsProperty == null) {
				throw new SpagicWorkflowException("Not Found The Process Instance Variable Name For Process ["
						+ processInstance.getProcess().getName() + "]");
			}
			return processInstanceStatusVariableNameAsProperty.getValue();
		} finally {
			if (aSession != null && aSession.isOpen())
				aSession.close();
		}
	}

	/**
	 * Gets the task instance status variable name.
	 * 
	 * @param taskName
	 *          the task name
	 * 
	 * @return the task instance status variable name
	 */
	public static String getTaskInstanceStatusVariableName(String processName, String taskName) {
		Session aSession = null;
		try {
			aSession = getSession();
			Process spagicProcess = AuditDBUtils.getLastDefinitionForProcess(aSession, processName);
			String taskNameInMetaDB = ProcessEngine.getStepLocalNamePart(spagicProcess.getName(), spagicProcess
					.getProcessVersion(), taskName);
			Step step = AuditDBUtils.getStepByName(aSession, taskNameInMetaDB);
			Property taskInstanceStatusVariableNameAsProperty = AuditDBUtils.getStepPropertyByName(aSession, step,
					"TaskInstanceStatusVariableName");
			if (taskInstanceStatusVariableNameAsProperty == null) {
				throw new SpagicWorkflowException("Not Found The Task Instance Variable Name For Task [" + taskName + "]");
			}
			return taskInstanceStatusVariableNameAsProperty.getValue();
		} finally {
			if (aSession != null && aSession.isOpen())
				aSession.close();
		}
	}

	/**
	 * Gets the task name by task id.
	 * 
	 * @param taskId
	 *          the task id
	 * 
	 * @return the task name by task id
	 */
	public static String getTaskNameByTaskId(long taskId) {
		//JbpmContext jbpmContext = jbpmConfiguration.createJbpmContext();
		JbpmContext jbpmContext = getJbpmContext();
		try {
			TaskInstance jbpmTaskInstance = jbpmContext.getTaskInstance(taskId);
			if (jbpmTaskInstance == null)
				throw new SpagicWorkflowException("Task Instance Not Found");
			return jbpmTaskInstance.getTask().getName();
		} finally {
			jbpmContext.close();
		}
	}

	//
	// STATIC METHOD ---- Used by SpagicMonitorService
	//
	/**
	 * Gets the step local name part.
	 * 
	 * @param processName
	 *          the process name
	 * @param processVersion
	 *          the process version
	 * @param stepName
	 *          the step name
	 * 
	 * @return the step local name part
	 */
	public static String getStepLocalNamePart(String processName, String processVersion, String stepName) {
		String localPart = processName + "." + stepName + VERSION_LABEL + processVersion;
		return localPart;
	}

	/**
	 * Gets the short task name.
	 * 
	 * @param completeName
	 *          the complete name
	 * 
	 * @return the short task name
	 */
	public static String getShortTaskName(String completeName) {
		int begin = completeName.indexOf(".") + 1;
		int end = completeName.lastIndexOf("_v");
		return completeName.substring(begin, end);
	}

	/**
	 * Creates the spagic correlation id.
	 * 
	 * @param jBPMProcessInstance
	 *          the j bpm process instance
	 * 
	 * @return the string
	 */
	public static String createSpagicCorrelationId(org.jbpm.graph.exe.ProcessInstance jBPMProcessInstance) {
		if (jBPMProcessInstance == null)
			return null;
		return JBPM_PREFIX + jBPMProcessInstance.getId();
	}

	/**
	 * Gets the last definition for process.
	 * 
	 * @param aSession
	 *          the a session
	 * @param processName
	 *          the process name
	 * 
	 * @return the last definition for process
	 */
	public static org.spagic.metadb.model.Process getLastDefinitionForProcess(Session aSession, String processName) {

		String lastVersion = AuditDBUtils.getCurrentProcessVersion(processName, aSession);
		org.spagic.metadb.model.Process metaDBProcess = AuditDBUtils.getProcessByNameAndVersion(aSession, processName,
				lastVersion);
		return metaDBProcess;

	}

	/**
	 * Gets the spagic process instance from jbpm process instance.
	 * 
	 * @param jBPMProcessInstance
	 *          the j bpm process instance
	 * @param aSession
	 *          the a session
	 * 
	 * @return the spagic process instance from jbpm process instance
	 */
	public static ProcessInstance getSpagicProcessInstanceFromJBPMProcessInstance(
			org.jbpm.graph.exe.ProcessInstance jBPMProcessInstance, Session aSession) {
		String correlationID = createSpagicCorrelationId(jBPMProcessInstance);
		ProcessInstance spagicProcessInstance = AuditDBUtils.getProcessInstanceByIdCorrelation(aSession, correlationID);
		return spagicProcessInstance;
	}

	/**
	 * Creates the process instance.
	 * 
	 * @param process
	 *          the process
	 * @param correlationID
	 *          the correlation id
	 * @param parentProcessInstance
	 *          the parent process instance
	 * @param aSession
	 *          the a session
	 * 
	 * @return the org.spagic.metadb.model. process instance
	 */
	public static org.spagic.metadb.model.ProcessInstance createProcessInstance(org.spagic.metadb.model.Process process,
			String correlationID, org.spagic.metadb.model.ProcessInstance parentProcessInstance, Session aSession) {
		org.spagic.metadb.model.ProcessInstance processInstance = new org.spagic.metadb.model.ProcessInstance();
		processInstance.setProcess(process);
		Date now = new Date();
		processInstance.setStartdate(now);
		processInstance.setEnddate(now);
		processInstance.setIdCorrelation(correlationID);
		processInstance.setState(ProcessInstanceStateConstants.PROCESS_ACTIVE);
		processInstance.setParentProcessInstance(parentProcessInstance);
		ProcessExecutionState pexState = getProcessExecutionState(ProcessExecutionStateConstants.NORMAL_EXECUTION);
		if (pexState == null)
			log.info("Process Execution State is Null");
		else
			log.info("Process Execution State is not null");
		processInstance.setProcessExecutionState(getProcessExecutionState(ProcessExecutionStateConstants.NORMAL_EXECUTION));
		Long idProcessSaved = (Long) aSession.save(processInstance);
		processInstance = (org.spagic.metadb.model.ProcessInstance) aSession.get(
				org.spagic.metadb.model.ProcessInstance.class, idProcessSaved);
		return processInstance;
	}

	

	/**
	 * Gets the process execution state.
	 * 
	 * @param processExecutionStateID
	 *          the process execution state id
	 * 
	 * @return the process execution state
	 */
	public static synchronized ProcessExecutionState getProcessExecutionState(long processExecutionStateID) {
		if (processExecutionStates == null) {
			loadProcessExecutionStates();
		}
		return processExecutionStates.get(processExecutionStateID);
	}

	/**
	 * Load process execution states.
	 */
	@SuppressWarnings("unchecked")
	private static synchronized void loadProcessExecutionStates() {
		Session aSession = null;
		try {
			processExecutionStates = new HashMap<Long, ProcessExecutionState>();
			aSession = HibernateUtil.getSessionFactory().openSession();
			List<ProcessExecutionState> ProcessExecutionStateList = aSession.createQuery(
					" from org.spagic.metadb.model.ProcessExecutionState").list();
			for (ProcessExecutionState pes : ProcessExecutionStateList) {
				processExecutionStates.put(pes.getIdProcessExecutionState(), pes);
			}
		} catch (Throwable t) {
			t.printStackTrace();
		} finally {
			aSession.close();
		}
	}

	
	
	public static boolean isSerializationNeeded(Object value){

		if (( value instanceof String )
			|| ( value instanceof Integer )
			|| ( value instanceof Long )
			|| ( value instanceof Double )
			|| ( value instanceof Date)){
			return false;
				
		}
		return true;
	}
	
	public static String serializeValue(Object value){
		XStream xstream = new XStream(new Dom4JDriver());
		return xstream.toXML(value);
	}

	public static  Object deserialize(Object streamed){
		XStream xStreamResObject = new XStream(new Dom4JDriver());
		Object valueObj = xStreamResObject.fromXML((String)streamed);
		return valueObj;
	}
	
	

	public static String createFlagVariable(String vName){
		return "__FLAGSERIALIZED_"+vName;	
	}

	protected static void storeVariable(String variableName, Object variableValue, org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance){
		Object toStoreValue = variableValue;
		if (ProcessEngine.isSerializationNeeded(variableValue)){
			toStoreValue = ProcessEngine.serializeValue(variableValue);
			jbpmProcessInstance.getContextInstance().setVariable(createFlagVariable(variableName), "TRUE");
		}
		jbpmProcessInstance.getContextInstance().setVariable(variableName, toStoreValue);
	}
	
	protected static Object getVariableValue(String varName, Object varValue, org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance){
		if (needToBeDeserialized(varName, jbpmProcessInstance))
			return ProcessEngine.deserialize(varValue);
		return varValue;
	}
	
	public static synchronized void signalToken(String tokenContextAsString, Variable[] vars) {
		
		TokenContext tokenContext = TokenContext.fromTokenContex(tokenContextAsString);
		Long rootTokenId = tokenContext.getRootTokenId();
		Long tokenId = tokenContext.getTokenId();
		
		ProcessEngine.lockToken(rootTokenId);
		JbpmContext jbpmContext = getJbpmContext();
		try {
			log.info("######["+Thread.currentThread().getId()+":"+Thread.currentThread().getName() +"] -> SIGNAL TOKEN START ["+tokenId+"]");
			Token token = jbpmContext.getTokenForUpdate(tokenId);
			
		
			if (token == null){
				log.error("######["+Thread.currentThread().getId()+":"+Thread.currentThread().getName() +"] -> TOKEN WITH ID  ["+tokenId+"] IS NULL");
			}else{
				log.info("$$ -> "+token.getNode().getName());
			}
			org.jbpm.graph.exe.ProcessInstance pi = token.getProcessInstance();
			if (pi == null){
				log.error("######["+Thread.currentThread().getId()+":"+Thread.currentThread().getName() +"] signalToken ["+tokenId+"] -> PROCESS INSTANCE IS NULL");
			}
			
			for (int i = 0; (vars != null && i < vars.length); i++) {
				pi.getContextInstance().setVariable(vars[i].getName(), vars[i].getValue(), token);
			}
			token.signal();
			
			log.info("######["+Thread.currentThread().getId()+":"+Thread.currentThread().getName() +"]-> SIGNAL TOKEN START ["+tokenId+"] OK ");;		
			
		} catch (Throwable t) {
			
			log.error("######["+Thread.currentThread().getId()+":"+Thread.currentThread().getName() +"] -> ERR [" + tokenId + "]", t);
			jbpmContext.setRollbackOnly();
		
		} finally {
	
			closeJbpmContextAndUnlockToken(jbpmContext, rootTokenId);
		
		}
	}
	
	public static synchronized void endWithException(String tokenContextAsString, Exception exception) {
		
		TokenContext tokenContext = TokenContext.fromTokenContex(tokenContextAsString);
		Long rootTokenId = tokenContext.getRootTokenId();
		Long tokenId = tokenContext.getTokenId();
		
		ProcessEngine.lockToken(rootTokenId);
		JbpmContext jbpmContext = getJbpmContext();
		try {
			Token token = jbpmContext.getTokenForUpdate(tokenId);
			org.jbpm.graph.exe.ProcessInstance pi = token.getProcessInstance();
			
			if (exception != null){
				FaultLog faultLog = new FaultLog();
				faultLog.setToken(token);
				faultLog.setAction(null);
				faultLog.setException(exception);
				faultLog.setProcessId(pi.getId());
				faultLog.setProcessName(pi.getProcessDefinition().getName());
				faultLog.setNodeName(token.getNode().getName());

				pi.getRootToken().addLog(faultLog);
			}
		} catch (Throwable t) {
			jbpmContext.setRollbackOnly();
			log.error("Error Signaling Token  [" + tokenId + "]", t);

		} finally {
			
			closeJbpmContextAndUnlockToken(jbpmContext, rootTokenId);
		
		}
	}

	public void setProfileService(IProfileService profileService) {
		this.profileService = profileService;
		
	}

	public IProfileService getProfileService() {
		return this.profileService;
	}
	
	
	public boolean isProfiled(TaskInstance taskIntance) {
		return ( taskIntance.getPooledActors() != null && taskIntance.getPooledActors().size() > 0);
	}
	
	public boolean isTaskAssignableToUser(TaskInstance taskIntance, String userId) {
		if (!isProfiled(taskIntance))
			return true;
		
		if (userId.equalsIgnoreCase(getTaskOwner(taskIntance)))
			return true;
		
		
		List<String> userRoles = profileService.getUserRoles(userId);
		Set<PooledActor> actorsEnabledForTaskInstance = taskIntance.getPooledActors();
		for (PooledActor pooledActor : actorsEnabledForTaskInstance){
			if (userRoles.contains(pooledActor.getActorId()))
					return true;
		}
		return false;
		
	}
	
	public String getTaskOwner(TaskInstance taskIntance) {
		return taskIntance.getActorId();
	}
	
	public void checkUserOnTask(TaskInstance taskIntance, String userId) {
		if (isSuperUser(userId)){
			return;
		}else{
		   String owner = getTaskOwner(taskIntance);
		   if ( owner != null ){
			   if (!userId.equalsIgnoreCase(owner))
				   throw new SpagicWorkflowException("The Task ["+taskIntance.getId()+"-"+taskIntance.getName()+"]  is owned by ["+owner+"] only this user could reject the task");
		    }else{
			   if (!isTaskAssignableToUser(taskIntance, userId))
				   throw new SpagicWorkflowException("User ["+userId+"] has not sufficient roles to work on  ["+taskIntance.getId()+"-"+taskIntance.getName()+"]");
			}
		}
	}

	public boolean isSuperUser(String userId) {
		return profileService.getUserRoles(userId).contains("su");
	}
	
	public org.spagic.workflow.api.TaskInstance getSpagicTaskInstanceFromJBPMProcessInstance(Session session, TaskInstance jbpmTaskInstance, org.spagic.workflow.api.Process processDef, long spagicProcessInstanceId, int completionPercentage) {
		org.spagic.workflow.api.TaskInstance ti = new org.spagic.workflow.api.TaskInstance();
	
		if (!jbpmTaskInstance.isOpen()) {
			if (jbpmTaskInstance.isCancelled()) {
				ti.setStatus(org.spagic.workflow.api.TaskInstance.CANCELLED);
			} else {
				ti.setStatus(org.spagic.workflow.api.TaskInstance.COMPLETED);
			}
		} else {
			if ( jbpmTaskInstance.getStart() != null ){
				ti.setStatus(org.spagic.workflow.api.TaskInstance.STARTED);
			}else{
				ti.setStatus(org.spagic.workflow.api.TaskInstance.GENERATED);
			}
		}
		Task taskDef = this.queryAPI.getHumanTask(processDef.getProcessId(), jbpmTaskInstance.getName());
		ti.setTaskName(jbpmTaskInstance.getName());
		ti.setTaskInstanceId(jbpmTaskInstance.getId());
		ti.setTaskId(taskDef.getTaskId());
		ti.setTaskInternalId(jbpmTaskInstance.getTask().getId());
		ti.setProcessInstanceId(spagicProcessInstanceId);
		ti.setCreationDate(jbpmTaskInstance.getCreate());
		ti.setEndDate(jbpmTaskInstance.getEnd());
		ti.setExpirationDate(jbpmTaskInstance.getDueDate());
		ti.setAcceptanceDate(jbpmTaskInstance.getStart());
		ti.setProcessName(processDef.getName());
		ti.setProcessId(processDef.getProcessId());					
		ti.setCompletionPercentage(completionPercentage);
		ti.setAssignedTo(jbpmTaskInstance.getActorId());
		
		String form = getFormForTaskInstance(session, jbpmTaskInstance);
		if (form != null)
			ti.setAssociatedFormName(form);
		return ti;
	}
	
	public org.spagic.workflow.api.TaskInstance getNotGeneratedSpagicTaskInstanceFromJBPMProcessInstance(Task staticTask, long spagicProcessInstanceId){
		org.spagic.workflow.api.TaskInstance ti = new org.spagic.workflow.api.TaskInstance();
		ti.setStatus(org.spagic.workflow.api.TaskInstance.NOT_GENERATED);
		ti.setTaskName(staticTask.getTaskName());
		ti.setTaskId(staticTask.getTaskId());
		ti.setTaskInternalId(staticTask
			.getInternalId());
		ti.setProcessInstanceId(spagicProcessInstanceId);
		ti.setCreationDate(null);
		ti.setEndDate(null);
		ti.setExpirationDate(null);
		ti.setAcceptanceDate(null);
		ti.setProcessName(staticTask.getProcessName());
		ti.setProcessId(staticTask.getProcessId());
		return ti;
	}
	
	public String getFormForTaskInstance(Session aSession, org.jbpm.taskmgmt.exe.TaskInstance jbpmTaskInstance){
		org.jbpm.graph.exe.ProcessInstance jbpmPi = jbpmTaskInstance.getProcessInstance();
		org.spagic.metadb.model.ProcessInstance spagicPi = ProcessEngine.getSpagicProcessInstanceFromJBPMProcessInstance(
				jbpmPi, aSession);
		
		String jbpmTaskName = jbpmTaskInstance.getTask().getName();
		org.spagic.metadb.model.Process spagicProcess = spagicPi.getProcess();
		String spagicProcessName = spagicProcess.getName();
		String processNamespace = spagicProcess.getNamespace();
		
		String stepName = ProcessEngine.getStepLocalNamePart(spagicProcessName, spagicProcess.getProcessVersion(),
				jbpmTaskName);
		Step step = AuditDBUtils.getStepByName(aSession, processNamespace, stepName);
		Property form = AuditDBUtils.getStepPropertyByName(aSession, step, "Form");
		if (form != null)
			return form.getValue();
		else 
			return null;

	}
	
	
	
	public static boolean needToBeDeserialized(String variableName, org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance){
		String v = (String)jbpmProcessInstance.getContextInstance().getVariable(createFlagVariable(variableName));
		return v != null && v.equalsIgnoreCase("TRUE");
	}
	
	
	public static boolean needToDeploy(final String processName, final Date currentDeployLastUpdatedTimeStamp ){
	
		Boolean result = (Boolean)doWithinSpagic("needToDeploy()", 
				null, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						Process process = getLastDefinitionForProcess(spagicSession, processName);
						if (process == null)
							return true; // The process is not deployed 
						
						// Process is deployed must check the date
						Date inDbDeployTimestamp = process.getStartdate();
						if (currentDeployLastUpdatedTimeStamp.after(inDbDeployTimestamp)){
							return true;
						}else{
							return false;
						}
					}
				}, null);
		
		return result;
	}
	
	public static final Long deployJdplFile(final String jpdlFileLocation){
		
			return (Long)doWithinJBPM("deployJdplFile", 
					null, new ISpagicJBPMCommand() {
						
						public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
								Session spagicSession,
								ProcessInstanceExecutionContext instanceExecutionContext) {
							
							File jpdlFile = new File(jpdlFileLocation);
							URL jpdlFileURL = null;
							try{
								jpdlFileURL = jpdlFile.toURI().toURL();
							
							
								ProcessDefinition processDefinition = ProcessDefinition.parseXmlInputStream(jpdlFileURL.openStream());
								jbpmContext.deployProcessDefinition(processDefinition);

								return new Long(processDefinition.getVersion());
								
							}catch (Exception e) {
								throw new RuntimeException(e);
							}
						}
					},null, false);
			
	}
	
	
	public static Date normalizeDate(Date d){
		try{
			String asString = sdf.format(d);
		
			Date normalizedDate = sdf.parse(asString);
			return normalizedDate;
			
		}catch (Exception e) {
			throw new RuntimeException(e.getMessage(), e);
		}
		

	}
}
