/**

    Copyright 2007, 2009 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
 **/
package org.spagic.workflow.api.jbpm;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.activation.DataHandler;
import javax.mail.util.ByteArrayDataSource;

import org.hibernate.Criteria;
import org.hibernate.Session;
import org.hibernate.criterion.Restrictions;
import org.jbpm.JbpmConfiguration;
import org.jbpm.JbpmContext;
import org.jbpm.graph.def.ProcessDefinition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic.graph.parser.SvgParser;
import org.spagic.metadata.SchemaDefinition;
import org.spagic.metadb.base.TypeServiceConstants;
import org.spagic.metadb.dbutils.AuditDBUtils;
import org.spagic.metadb.model.ProcessDiagram;
import org.spagic.metadb.model.Property;
import org.spagic.metadb.model.SchemaMetadata;
import org.spagic.metadb.model.Step;
import org.spagic.metadb.model.SubProcess;
import org.spagic.metadb.model.Transition;
import org.spagic.metadb.model.TransitionInstance;
import org.spagic.workflow.api.IProcessEngine;
import org.spagic.workflow.api.IQueryAPI;
import org.spagic.workflow.api.Process;
import org.spagic.workflow.api.ProcessInstance;
import org.spagic.workflow.api.SpagicWorkflowException;
import org.spagic.workflow.api.SubProcessTask;
import org.spagic.workflow.api.Task;
import org.spagic.workflow.api.TaskInstance;
import org.spagic.workflow.api.Variable;

/**
 * The Class QueryAPI.
 */
public class QueryAPI implements IQueryAPI {

	/** The Constant log. */
	private static Logger log = LoggerFactory.getLogger(QueryAPI.class);

	/** The jbpm configuration. */
	private JbpmConfiguration jbpmConfiguration = null;

	/** The Constant OUTGOING_CONDITION_VALUE_PREFIX. */
	private static final String OUTGOING_CONDITION_VALUE_PREFIX = "OutGoingValue";

	public static final String PROPERTY_FORM_ID = "FormId";
	public static final String PROPERTY_FORM_TYPE = "FormType";
	public static final String PROPERTY_FORM_TITLE = "FormTitle";

	public static final String PROPERTY_FORM_APPLICATION = "FormApplication";
	public static final String PROPERTY_FORM_FORM = "FormForm";

	private static final String PROPERTY_FORM_TASK = "FormTask";

	/** The jbpm engine. */
	private ProcessEngine jbpmEngine = null;

	/**
	 * Instantiates a new query api.
	 * 
	 * @param jbpmConfiguration
	 *          the jbpm configuration
	 * @param processEngine
	 *          the process engine
	 */
	public QueryAPI(JbpmConfiguration jbpmConfiguration, IProcessEngine processEngine) {
		super();
		this.jbpmConfiguration = jbpmConfiguration;
		this.jbpmEngine = (ProcessEngine) processEngine;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getHumanTask(long)
	 */
	public Task[] getHumanTask(long processId) {
		Process aProcess = getProcess(processId);
		if (aProcess == null) {
			throw new SpagicWorkflowException(" Process with ID [" + processId + "] Not Found");
		}
		return getHumanTask(aProcess);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getHumanTask(java.lang.String)
	 */
	public Task[] getHumanTask(String processName) {
		Process aProcess = getProcess(processName);
		if (aProcess == null) {
			throw new SpagicWorkflowException(" Process with Name [" + processName + "] Not Found");
		}
		return getHumanTask(aProcess);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.spagic.workflow.api.IQueryAPI#getHumanTask(org.spagic.workflow.api.
	 * Process)
	 */
	public Task[] getHumanTask(final Process process) {
		
		
		return (Task[])ProcessEngine.doQueryWithinJBPMAndSpagic("QueryAPI.getHumanTask()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						ProcessDefinition def = null;

						int jbpmVersion = 0;
						try {
							jbpmVersion = (new Integer(process.getVersion()).intValue());
						} catch (Exception e) {
						}
						if (jbpmVersion == 0) {
							def = jbpmContext.getGraphSession().findLatestProcessDefinition(process.getName());
						} else {
							def = jbpmContext.getGraphSession().findProcessDefinition(process.getName(), jbpmVersion);
						}
						
						if (def == null) {
							log.warn("Process [" + process.getName() + "] Version[" + process.getVersion()
									+ "] is not a Spagic Workflow process");
							return new Task[0];
						}
						
						// Retrieve the spagic process
						org.spagic.metadb.model.Process spagicProcess = AuditDBUtils.getProcessByNameAndVersion(spagicSession, process
								.getName(), process.getVersion());

						// Retrieve all JBPM task for this process
						Map<String, org.jbpm.taskmgmt.def.Task> jbpmTasks = def.getTaskMgmtDefinition().getTasks();

						Task[] tasks = new Task[jbpmTasks.size()];
						int i = 0;
						for (String str : jbpmTasks.keySet()) {
							org.jbpm.taskmgmt.def.Task jbpmTask = jbpmTasks.get(str);
							tasks[i] = new Task();

							Step step = AuditDBUtils.getNamedStepForProcess(spagicSession, spagicProcess.getIdProcess(), ProcessEngine
									.getStepLocalNamePart(process.getName(), process.getVersion(), jbpmTask.getName()));

							tasks[i].setTaskId(step.getIdStep());
							tasks[i].setInternalId(jbpmTask.getId());
							tasks[i].setTaskName(jbpmTask.getName());
							String pooledActorsExpression = jbpmTask.getPooledActorsExpression();
							if (pooledActorsExpression != null){
								String[] roles = pooledActorsExpression.split(",");
								tasks[i].setEnabledRoles(Arrays.asList(roles));
							}
							
							tasks[i].setProcessName(spagicProcess.getName());
							tasks[i].setProcessId(spagicProcess.getIdProcess());
							i++;
						}
						return tasks;
					}
				}, null);

	}

	public TaskInstance[] getHumanTaskInstances(long processInstanceId) {
		return getHumanTaskInstances(processInstanceId,null);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getTaskForActor(long,
	 * java.lang.String)
	 */
	public TaskInstance[] getHumanTaskInstancesForActorAndProcess(long processInstanceId, String actorID) {
		
		return getHumanTaskInstances(processInstanceId,actorID);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getTaskForActor(long,
	 * java.lang.String)
	 */
	public TaskInstance[] getHumanTaskInstancesForActor(final String actorID) {
		if (jbpmEngine.isSuperUser(actorID))
			return getGeneratedHumanTaskInstances();
		
		return (TaskInstance[])ProcessEngine.doQueryWithinJBPMAndSpagic("QueryAPI.getHumanTaskInstancesForActor()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						List<String> rolesForActor = jbpmEngine.getProfileService().getUserRoles(actorID);
						
						
						List<TaskInstance> taskInstances = new ArrayList<TaskInstance>();
				
						if (!rolesForActor.isEmpty()){
							Collection<org.jbpm.taskmgmt.exe.TaskInstance> jbpmTaskInstances = null;
							jbpmTaskInstances = jbpmContext.getTaskMgmtSession().findPooledTaskInstances(rolesForActor);
							
							
							for (org.jbpm.taskmgmt.exe.TaskInstance jbpmTaskInstance : jbpmTaskInstances ) {
								org.spagic.metadb.model.ProcessInstance spagicPi = jbpmEngine.getSpagicProcessInstanceFromJBPMProcessInstance(jbpmTaskInstance.getProcessInstance(), spagicSession); 
								String processName = jbpmTaskInstance.getProcessInstance().getProcessDefinition().getName();
								Process process = getProcess(processName);
								TaskInstance ti = jbpmEngine.getSpagicTaskInstanceFromJBPMProcessInstance(spagicSession,jbpmTaskInstance, process, spagicPi.getIdProcessInstance(), getTaskCompletionPercentage(jbpmTaskInstance));
							
								taskInstances.add(ti);
							
							}
							
							
						}	
						
						TaskInstance[] returnedTaskInstances = new TaskInstance[taskInstances.size()];
						returnedTaskInstances = taskInstances.toArray(returnedTaskInstances);
						return returnedTaskInstances;
					}
			},null);
		
		
	}
	
	
	
	
	private boolean checkIsAssigned(TaskInstance t, String actorID) {
		boolean result = false;
		if(t.getStatus()==TaskInstance.STARTED &&
				(t.getAssignedTo()).equals(actorID))
			result = true;
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getHumanTaskInstances(long) This
	 * function uses only the JBPM database
	 */
	@SuppressWarnings("unchecked")
	public TaskInstance[] getHumanTaskInstances(final long processInstanceId, final String actorID) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		return (TaskInstance[])ProcessEngine.doQueryWithinJBPMAndSpagic("QueryAPI.getHumanTaskInstances()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
			
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.loadProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
						String processName = jbpmProcessInstance.getProcessDefinition().getName();
			
						Process processDef = getProcess(processName);
						List<TaskInstance> taskInstances = new ArrayList<TaskInstance>();
						List<String> rolesForActor = new ArrayList<String>();
						
						// Retrieves all human tasks defined for process
						List<String> addedTask = new ArrayList<String>();
						
						if (jbpmEngine.isSuperUser(actorID)) {
			//				return getGeneratedHumanTaskInstances(null, null, null, null, null, null, 0, processInstanceId);
							Collection<org.jbpm.taskmgmt.exe.TaskInstance> jbpmTaskInstances = null;
			
			
							jbpmTaskInstances = jbpmProcessInstance.getTaskMgmtInstance().getTaskInstances();
							
							// Add all generated task
							for (org.jbpm.taskmgmt.exe.TaskInstance jbpmTaskInstance : jbpmTaskInstances ) {
							    if (actorID != null){
							    	if (!jbpmEngine.isTaskAssignableToUser(jbpmTaskInstance, actorID))
							    		continue;
							    }
							    
							    TaskInstance ti = jbpmEngine.getSpagicTaskInstanceFromJBPMProcessInstance(spagicSession,jbpmTaskInstance, processDef, processInstanceId, getTaskCompletionPercentage(jbpmTaskInstance));
								taskInstances.add(ti);
							}
			
						} 
						else if (actorID != null){
						//
						// IF ACTOR ID != NULL 
						// 1 - Retrieve roles
						// 2 - Get Only my assignable task
						//
							rolesForActor = jbpmEngine.getProfileService().getUserRoles(actorID);
							if (!rolesForActor.isEmpty()){
								Collection<org.jbpm.taskmgmt.exe.TaskInstance> jbpmTaskInstances = null;
							
								jbpmTaskInstances = jbpmProcessInstance.getTaskMgmtInstance().getTaskInstances();
								
								// Add all generated task
								for (org.jbpm.taskmgmt.exe.TaskInstance jbpmTaskInstance : jbpmTaskInstances ) {
								    if (actorID != null){
								    	if (!jbpmEngine.isTaskAssignableToUser(jbpmTaskInstance, actorID))
								    		continue;
								    }
								    
								    TaskInstance ti = jbpmEngine.getSpagicTaskInstanceFromJBPMProcessInstance(spagicSession,jbpmTaskInstance, processDef, processInstanceId, getTaskCompletionPercentage(jbpmTaskInstance));
									taskInstances.add(ti);
									addedTask.add(ti.getTaskName());
								}
				
							}
						}
						else {
							// we have NO actor, first we get the generated task and then the non 
							// generated ones.
							
							// First of all, get generated tasks.
							TaskInstance staticTasksInstances[] = getGeneratedHumanTaskInstances(processInstanceId);
							if(staticTasksInstances != null) {
								int max = staticTasksInstances.length;	
								// Add as NOT_GENERATED all task defined that are not been added before
								// Retrieves all human tasks defined for process
								for (int i = 0; i < max; i++) {
									if (!addedTask.contains(staticTasksInstances[i].getTaskName())) {
										taskInstances.add(staticTasksInstances[i]);
										addedTask.add(staticTasksInstances[i].getTaskName());
									}
								}
							}
							// NOT GENERATED TASK DOES NOT MAKE SENSE IF actor i specified
							Task[] staticTasksForProcess = getHumanTask(processName);
							if (staticTasksForProcess != null) {
							int max = staticTasksForProcess.length;
							for (int i = 0; i < max; i++) {
								if (!addedTask.contains(staticTasksForProcess[i].getTaskName())) {
									TaskInstance ti = jbpmEngine.getNotGeneratedSpagicTaskInstanceFromJBPMProcessInstance(staticTasksForProcess[i], processInstanceId);
									taskInstances.add(ti);
									addedTask.add(staticTasksForProcess[i].getTaskName());
								}
							}
							}
						}
						TaskInstance[] returnedTaskInstances = new TaskInstance[taskInstances.size()];
						returnedTaskInstances = taskInstances.toArray(returnedTaskInstances);
						return returnedTaskInstances;
					}
		},iec);

	}

	private int getTaskCompletionPercentage(org.jbpm.taskmgmt.exe.TaskInstance jbpmTaskInstance) {
		List<org.jbpm.graph.exe.Comment> comments = jbpmTaskInstance.getComments();

		// Retrieve the percentage comment and delete it
		for (Iterator<org.jbpm.graph.exe.Comment> it = comments.iterator(); it.hasNext();) {
			org.jbpm.graph.exe.Comment comment = it.next();
			if (comment.getMessage().startsWith(ProcessEngine.COMPLETION_PERCENTAGE_COMMENT)) {
				String[] parts = comment.getMessage().split(ProcessEngine.COMMENT_SEPARATION);
				if (parts.length == 2) {
					return Integer.valueOf(parts[1]);
				}
			}
		}

		return 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getHumanTaskInstances(long,
	 * java.lang.String, java.lang.String)
	 */
	public TaskInstance[] getHumanTaskInstances(final long processInstanceId, final String attributeName, final String attributeValue) {
				
		final TaskInstance[] ti = getHumanTaskInstances(processInstanceId);
		final String processName = getProcessInstanceById(processInstanceId).getProcess().getName();
		
		return (TaskInstance[])ProcessEngine.doQueryWithinJBPMAndSpagic("QueryAPI.getHumanTaskInstances()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						List<TaskInstance> taskInstances = null;
						
						org.spagic.metadb.model.Process spagicProcess = ProcessEngine.getLastDefinitionForProcess(spagicSession, processName);

						taskInstances = new ArrayList<TaskInstance>();
						for (int i = 0; i < ti.length; i++) {
							Step step = AuditDBUtils.getNamedStepForProcess(spagicSession, spagicProcess.getIdProcess(), ProcessEngine
									.getStepLocalNamePart(processName, spagicProcess.getProcessVersion(), ti[i].getTaskName()));
							List<Property> props = AuditDBUtils.getStepPropertyWithNameLike(spagicSession, step, "TaskAttributes");
							for (Property aProperty : props) {
								String code = aProperty.getCode();
								String name = null;
								String value = null;
								if (code.endsWith(".name")) {
									name = aProperty.getValue();
									for (Property aPropertyName : props) {
										String codeVal = aPropertyName.getCode();
										if (codeVal.startsWith(code.substring(0, code.lastIndexOf("."))) && codeVal.endsWith(".value"))
											value = aPropertyName.getValue();
									}
									if (name != null && value != null && name.equals(attributeName) && value.equals(attributeValue)) {
										taskInstances.add(ti[i]);
									}
								}
							}
						}

						TaskInstance[] toReturn = new TaskInstance[taskInstances.size()];
						toReturn = taskInstances.toArray(toReturn);

						return toReturn;	
					}
		}, null);
		
	}

	

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getProcess(long)
	 */
	public Process getProcess(final long processId) {
		return (Process)ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcess()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						return getProcess(aSession, processId);
					}
				},null); 
			
		
	}

	public Process getProcess(Session aSession, long processId) {
		org.spagic.metadb.model.Process spagicMetaDBProcess = AuditDBUtils.getProcessById(aSession, processId);
		if (spagicMetaDBProcess == null) {
			throw new SpagicWorkflowException(" Process with ID [" + processId + "] Not Found");
		}
		return getProcessBO(spagicMetaDBProcess);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getProcess(java.lang.String)
	 */
	public Process getProcess(final String processName) {
		
		return (Process)ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcess()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						org.spagic.metadb.model.Process spagicMetaDBProcess = ProcessEngine.getLastDefinitionForProcess(aSession,
								processName);
						if (spagicMetaDBProcess == null) {
							throw new SpagicWorkflowException(" Process with Name [" + processName + "] Not Found");
						}
						return getProcessBO(spagicMetaDBProcess);
					}
				},null); 
		
	}
	

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getProcessGraph(long,
	 * java.lang.String)
	 */
	public String getProcessGraph(final long processId) {
		
		return (String)ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcessGraph()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.Process process = AuditDBUtils.getProcessById(aSession, processId);
						ProcessDiagram processDiagram = process.getProcessDiagrams().iterator().next();

						String svgString = processDiagram.getSvgDiag();
						try{
							SvgParser parser = new SvgParser(processId, svgString, new ProcessInfo(aSession));
							return parser.createProcessSvgString();
						}catch (Exception e) {
							throw new SpagicWorkflowException(e);
						}
					}
				},null); 
		
	}
	

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getProcessInstanceById(long)
	 * Retrieves the Process instance by id.
	 */
	public ProcessInstance getProcessInstanceById(final long processInstanceId) {
		
		return (ProcessInstance)ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcessInstanceById()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.ProcessInstance spagicProcessInstance = AuditDBUtils.getProcessInstanceById(aSession,
								processInstanceId);
						if (spagicProcessInstance == null) {
							throw new SpagicWorkflowException(" ProcessInstance with ID [" + processInstanceId + "] Not Found");
						}
						return getProcessInstanceBO(spagicProcessInstance);
					}
				},null); 
	
	}

	/**
	 * Produces a JBPM process instance from a Spagic process instance.
	 * 
	 * @param spagicProcessInstance
	 *          the spagic process instance
	 * 
	 * @return A JBPM process instance
	 */
	public ProcessInstance getProcessInstanceBO(org.spagic.metadb.model.ProcessInstance spagicProcessInstance) {
		ProcessInstance processInstance = new ProcessInstance();

		processInstance.setName(spagicProcessInstance.getProcess().getName());
		processInstance.setProcessInstanceId(spagicProcessInstance.getIdProcessInstance());
		processInstance.setStart(spagicProcessInstance.getStartdate());
		processInstance.setEnd(spagicProcessInstance.getEnddate());
		processInstance.setState(spagicProcessInstance.getState());
		processInstance.setProcess(getProcessBO(spagicProcessInstance.getProcess()));

		return processInstance;
	}

	/**
	 * Produces a JBPM process from a Spagic process.
	 * 
	 * @param spagicProcess
	 *          the spagic process
	 * 
	 * @return A JBPM process
	 */
	public Process getProcessBO(org.spagic.metadb.model.Process spagicProcess) {
		Process process = new Process();
		setProcessBO(spagicProcess, process);
		return process;
	}

	/**
	 * Fill all the fields of a JBPM process retrieving the values from a Spagic
	 * process.
	 * 
	 * @param spagicProcess
	 *          the spagic process
	 * @param process
	 *          the process
	 */
	public void setProcessBO(org.spagic.metadb.model.Process spagicProcess, Process process) {
		process.setProcessId(spagicProcess.getIdProcess());
		process.setName(spagicProcess.getName());
		process.setVersion(spagicProcess.getProcessVersion());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getProcessInstanceGraph(long,
	 * java.lang.String)
	 */
	public String getProcessInstanceGraph(final long processInstanceId) {
		return (String)ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcessInstanceGraph()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						org.spagic.metadb.model.ProcessInstance processInstance = AuditDBUtils.getProcessInstanceById(aSession,
								processInstanceId);
						org.spagic.metadb.model.Process process = processInstance.getProcess();
						ProcessDiagram processDiagram = process.getProcessDiagrams().iterator().next();
						String svgString = processDiagram.getSvgDiag();
						try{
							SvgParser parser = new SvgParser(process.getIdProcess(), new Long(processInstanceId), svgString, new ProcessInfo(
								aSession));
							return parser.createProcessInstanceSvgString();
						}catch (Exception e) {
							throw new SpagicWorkflowException(e);
						}
						
					}
				},null); 
		
	}
		


	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getProcesses() Retrieves all
	 * processes.
	 */
	public Process[] getProcesses() {
		return (Process[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcesses()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						List<org.spagic.metadb.model.Process> spagicMetaDBProcesses = AuditDBUtils.getAllProcesses(aSession);

						Process[] toReturnProcesses = new Process[spagicMetaDBProcesses.size()];

						int i = 0;
						for (org.spagic.metadb.model.Process spagicProcess : spagicMetaDBProcesses) {
							toReturnProcesses[i] = getProcessBO(spagicProcess);
							i++;
						}
						return toReturnProcesses;
						
					}
				},null); 
		
		
		
	}
	


	/**
	 * Gets the processes.
	 * 
	 * @param onlyStandalone
	 *          the only standalone
	 * 
	 * @return the processes
	 * 
	 *         Retrieves all processes that can be activated only as main
	 *         processes, and not as subprocesses.
	 */
	public Process[] getProcesses(boolean onlyStandalone) {
		if (onlyStandalone == false) {
			return getProcesses();
		} else {
			return getProcessesByFilter("IsStandalone", "true", false);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.spagic.workflow.api.IQueryAPI#getProcessesByFilter(java.lang.String,
	 * java.lang.String, boolean)
	 */
	public Process[] getProcessesByFilter(final String attributeName, final String attributeValue, final boolean onlyStandalone) {

		return (Process[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcessesByFilter())", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						List<org.spagic.metadb.model.Process> spagicMetaDBProcesses = AuditDBUtils.getProcessesFilteredByPropertyName(
								aSession, attributeName, attributeValue);

						List<Process> toReturnProcesses = new ArrayList<Process>();

						for (org.spagic.metadb.model.Process spagicProcess : spagicMetaDBProcesses) {
							if (onlyStandalone) {
								Property standaloneProperty = AuditDBUtils.getProcessProperty(aSession, spagicProcess, "IsStandalone");
								if (standaloneProperty != null && standaloneProperty.getValue().equalsIgnoreCase("true")) {
									toReturnProcesses.add(getProcessBO(spagicProcess));

								}
							} else {
								toReturnProcesses.add(getProcessBO(spagicProcess));

							}
						}
						Process[] toReturnProcessesArray = new Process[toReturnProcesses.size()];
						return (Process[]) toReturnProcesses.toArray(toReturnProcessesArray);
						
					}
				},null); 
	}
	

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getLatestDefinitionProcesses()
	 */
	public Process[] getLatestDefinitionProcesses() {
		return (Process[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getLatestDefinitionProcesses()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						
						List<org.spagic.metadb.model.Process> spagicMetaDBProcesses = AuditDBUtils
						.getAllLatestDefinitionProcesses(aSession);

						Process[] toReturnProcesses = new Process[spagicMetaDBProcesses.size()];

						int i = 0;
						for (org.spagic.metadb.model.Process spagicProcess : spagicMetaDBProcesses) {
							toReturnProcesses[i] = getProcessBO(spagicProcess);
							i++;
						}
						return toReturnProcesses;
					}
				},null); 
	}


	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.spagic.workflow.api.IQueryAPI#getLatestDefinitionProcesses(boolean)
	 */
	public Process[] getLatestDefinitionProcesses(boolean onlyStandalone) {
		if (onlyStandalone == false) {
			return getLatestDefinitionProcesses();
		} else {
			return getLatestDefinitionProcessesByFilter("IsStandalone", "true", false);
		}
	}

	/**
	 * 
	 * @see
	 * org.spagic.workflow.api.IQueryAPI#getLatestDefinitionProcessesByFilter(
	 * java.lang.String, java.lang.String, boolean)
	 */
	public Process[] getLatestDefinitionProcessesByFilter(final String attributeName, final String attributeValue,
			final boolean onlyStandalone) {
		
		return (Process[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getLatestDefinitionProcessesByFilter()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						List<org.spagic.metadb.model.Process> spagicMetaDBProcesses = AuditDBUtils.getAllLatestDefinitionProcesses(aSession);

						List<Process> toReturnProcesses = new ArrayList<Process>();
		
						for (org.spagic.metadb.model.Process spagicProcess : spagicMetaDBProcesses) {
		
							// Verify the
							Property attributeMatchProperty = AuditDBUtils.getProcessProperty(aSession, spagicProcess, attributeName);
		
							if (attributeMatchProperty != null && attributeMatchProperty.getValue().equalsIgnoreCase(attributeValue)) {
		
								if (onlyStandalone) {
									Property standaloneProperty = AuditDBUtils.getProcessProperty(aSession, spagicProcess, "IsStandalone");
									if (standaloneProperty != null && standaloneProperty.getValue().equalsIgnoreCase("true")) {
										toReturnProcesses.add(getProcessBO(spagicProcess));
		
									}
								} else {
									toReturnProcesses.add(getProcessBO(spagicProcess));
								}
		
							}
		
						}
						Process[] toReturnProcessesArray = new Process[toReturnProcesses.size()];
						return (Process[]) toReturnProcesses.toArray(toReturnProcessesArray);
						
					}
				},null); 
	}

	
			


	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getSubProcessInstances(long)
	 * Retrieves all sub-process instances given the id of the parent process.
	 */
	public ProcessInstance[] getSubProcessInstances(final long parentProcessInstanceId) {
		return (ProcessInstance[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getSubProcessInstances()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.ProcessInstance parentProcessInstance = AuditDBUtils.getProcessInstanceById(aSession,
								parentProcessInstanceId);
						Set<org.spagic.metadb.model.ProcessInstance> childProcesses = parentProcessInstance.getChildProcessInstances();

						ProcessInstance[] result = new ProcessInstance[childProcesses.size()];
						int index = 0;
						for (org.spagic.metadb.model.ProcessInstance childProcess : childProcesses) {
							result[index++] = getProcessInstanceBO(childProcess);
						}

						return result;
						
					}
				},null); 
	}


	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getSubProcesses(long) Retrieves all
	 * processes that are used as subprocesses within the process identified by
	 * processId.
	 */
	public Process[] getSubProcesses(final long processId) {
		return (Process[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getSubProcesses()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.Process spagicMetaDBProcess = AuditDBUtils.getProcessById(aSession, processId);
						if (spagicMetaDBProcess == null) {
							throw new SpagicWorkflowException(" Process with ID [" + processId + "] Not Found");
						}

						Process[] result = getSubProcesses(spagicMetaDBProcess);
						return result;
						
					}
				},null); 
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getSubProcesses(java.lang.String)
	 * Retrieves the subprocesses used in a main process, given its name.
	 */
	public Process[] getSubProcesses(final String processName) {
		return (Process[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getSubProcesses()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.Process spagicMetaDBProcess = ProcessEngine.getLastDefinitionForProcess(aSession,
								processName);
						if (spagicMetaDBProcess == null) {
							throw new SpagicWorkflowException(" Process with name [" + processName + "] Not Found");
						}

						Process[] result = getSubProcesses(spagicMetaDBProcess);
						return result;
					}
				},null); 
	}

	/**
	 * Gets the sub processes.
	 * 
	 * @param spagicMetaDBProcess
	 *          the spagic meta db process
	 * 
	 * @return the sub processes
	 */
	public Process[] getSubProcesses(org.spagic.metadb.model.Process spagicMetaDBProcess) {
		Set<SubProcess> subProcesses = spagicMetaDBProcess.getSubProcessesForIdProcess();

		List<Process> toReturnProcesses = new ArrayList<Process>();

		for (SubProcess subProcess : subProcesses) {
			Process process = getProcessBO(subProcess.getProcessByIdSubprocess());
			if (!toReturnProcesses.contains(process))
				toReturnProcesses.add(process);
		}
		Process[] toReturnProcessesArray = new Process[toReturnProcesses.size()];
		return (Process[]) toReturnProcesses.toArray(toReturnProcessesArray);
	}

	/**
	 * Find task instance for task.
	 * 
	 * @param taskInstances
	 *          the task instances
	 * @param taskName
	 *          the task name
	 * 
	 * @return the org.jbpm.taskmgmt.exe. task instance
	 */
	protected org.jbpm.taskmgmt.exe.TaskInstance findTaskInstanceForTask(
			Collection<org.jbpm.taskmgmt.exe.TaskInstance> taskInstances, String taskName) {

		for (org.jbpm.taskmgmt.exe.TaskInstance ti : taskInstances) {
			if (ti.getName().equals(taskName)) {
				return ti;

			}
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.spagic.workflow.api.IQueryAPI#getHumanTaskOutgoingValues(java.lang.
	 * String, java.lang.String)
	 */
	public String[] getHumanTaskOutgoingValues(final String processName, final String taskName) {
		return (String[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getHumanTaskOutgoingValues()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.Process spagicProcess = ProcessEngine.getLastDefinitionForProcess(aSession, processName);
						Step step = AuditDBUtils.getNamedStepForProcess(aSession, spagicProcess.getIdProcess(), ProcessEngine
								.getStepLocalNamePart(processName, spagicProcess.getProcessVersion(), taskName));

						return getStepOutgoingValues(aSession, step);
					}
				},null); 
	}
	

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getHumanTaskOutgoingValues(long,
	 * java.lang.String)
	 */
	public String[] getHumanTaskOutgoingValues(final long processId, final String taskName) {
		return (String[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getHumanTaskOutgoingValues()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.Process spagicProcess = AuditDBUtils.getProcessById(aSession, processId);
						Step step = AuditDBUtils.getNamedStepForProcess(aSession, spagicProcess.getIdProcess(), ProcessEngine
								.getStepLocalNamePart(spagicProcess.getName(), spagicProcess.getProcessVersion(), taskName));

						return getStepOutgoingValues(aSession, step);
					}
				},null); 
	}


	/**
	 * Gets the decision step for step.
	 * 
	 * @param aSession
	 *          the a session
	 * @param aStep
	 *          the a step
	 * 
	 * @return the decision step for step
	 */
	protected Step getDecisionStepForStep(Session aSession, Step aStep) {

		Property property = AuditDBUtils.getStepPropertyByName(aSession, aStep, "im.servicebindingname");

		if (property.getValue().equalsIgnoreCase("Decision")) {
			// This step is a decision step return
			return aStep;
		} else {
			Set<Transition> outgoingTransitions = aStep.getTransitionsForIdStepSource();

			if (outgoingTransitions.size() > 0) {
				Step decisionStep = outgoingTransitions.iterator().next().getStepByIdStepTarget();
				return decisionStep;
			}
			return null;
		}
	}

	/**
	 * Gets the step outgoing values.
	 * 
	 * @param aSession
	 *          the a session
	 * @param step
	 *          the step
	 * 
	 * @return the step outgoing values
	 */
	protected String[] getStepOutgoingValues(Session aSession, Step step) {

		//
		// Find the decision node
		//

		String[] returnValues = new String[0];

		Step decisionStep = getDecisionStepForStep(aSession, step);

		if (decisionStep != null) {
			List<Property> outgoingValues = AuditDBUtils.getStepPropertyWithNameLike(aSession, decisionStep,
					OUTGOING_CONDITION_VALUE_PREFIX);
			returnValues = new String[outgoingValues.size()];

			int k = 0;
			for (Property aProperty : outgoingValues) {
				returnValues[k] = aProperty.getValue();
				k++;
			}
		}

		return returnValues;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getProcessInstances(long)
	 */
	public ProcessInstance[] getProcessInstances(final long processId) {
		return (ProcessInstance[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcessInstances()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						List<org.spagic.metadb.model.ProcessInstance> instances = AuditDBUtils.getProcessInstancesForProcess(aSession,
								processId);

						ProcessInstance[] toReturn = new ProcessInstance[instances.size()];
						int i = 0;

						for (org.spagic.metadb.model.ProcessInstance pi : instances) {
							toReturn[i] = getProcessInstanceBO(pi);
							i++;
						}
						return toReturn;
					}
				},null); 
		
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.spagic.workflow.api.IQueryAPI#getProcessInstances(java.lang.String)
	 */
	public ProcessInstance[] getProcessInstances(final String processName) {
		return (ProcessInstance[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcessInstances()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.Process spagicProcess = ProcessEngine.getLastDefinitionForProcess(aSession, processName);
						List<org.spagic.metadb.model.ProcessInstance> instances = AuditDBUtils.getProcessInstancesForProcess(aSession,
								spagicProcess.getIdProcess());

						ProcessInstance[] toReturn = new ProcessInstance[instances.size()];
						int i = 0;

						for (org.spagic.metadb.model.ProcessInstance pi : instances) {
							toReturn[i] = getProcessInstanceBO(pi);
							i++;
						}
						return toReturn;
					}
				},null); 

		
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getProcessInstances()
	 */
	public ProcessInstance[] getProcessInstances() {
		return (ProcessInstance[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcessInstances()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						List<org.spagic.metadb.model.ProcessInstance> instances = AuditDBUtils.getAllProcessInstancesByTechnology(
								aSession, TypeServiceConstants.JBPM_PROCESS);

						ProcessInstance[] toReturn = new ProcessInstance[instances.size()];
						int i = 0;

						for (org.spagic.metadb.model.ProcessInstance pi : instances) {
							toReturn[i] = getProcessInstanceBO(pi);
							i++;
						}
						return toReturn;
					}
				},null); 
	}
	
	
			
		

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getLinkedProcessInstances(long)
	 */
	public ProcessInstance[] getLinkedProcessInstances(final long processInstanceId) {
		return (ProcessInstance[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getLinkedProcessInstances()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.ProcessInstance processInstance = AuditDBUtils.getProcessInstanceById(aSession,
								processInstanceId);

						if (processInstance == null) {
							throw new SpagicWorkflowException(" Process instance with ID [" + processInstanceId + "] Not Found");
						}

						List<org.spagic.metadb.model.ProcessInstance> processInstances = AuditDBUtils.getProcessInstancesByIdPool(
								aSession, processInstance.getPoolInstance().getIdPool());
						ProcessInstance[] result = new ProcessInstance[processInstances.size()];
						int index = 0;
						for (org.spagic.metadb.model.ProcessInstance currentProcessInstance : processInstances) {
							result[index++] = getProcessInstanceBO(currentProcessInstance);
						}

						return result;
					}
				},null); 
	}

	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getProcessInstanceStatus(long)
	 */
	public String getProcessInstanceStatus(long processInstanceId) {
		String processInstanceStatusVariableName = ProcessEngine.getProcessInstanceStatusVariableName(processInstanceId);
		return String.valueOf(jbpmEngine.getControlAPI().getGlobalVariable(processInstanceId,
				processInstanceStatusVariableName).getValue());
	}

	/**
	 * Gets the task instance status.
	 * 
	 * @param processInstanceId
	 *          the process instance id
	 * @param taskName
	 *          the task name
	 * 
	 * @return the task instance status
	 */
	public String getTaskInstanceStatus(long processInstanceId, String taskName) {
		ProcessInstance pi = getProcessInstanceById(processInstanceId);
		String taskStatusVariableName = ProcessEngine.getTaskInstanceStatusVariableName(pi.getName(), taskName);
		return String.valueOf(jbpmEngine.getControlAPI().getGlobalVariable(processInstanceId, taskStatusVariableName));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getTaskInstanceStatus(long, long)
	 */
	public String getTaskInstanceStatus(long processInstanceId, long taskInstanceId) {
		String taskName = ProcessEngine.getTaskNameByTaskId(taskInstanceId);
		return getTaskInstanceStatus(processInstanceId, taskName);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getSubprocessTasks(long)
	 */
	public SubProcessTask[] getSubprocessTasks(final long processId) {
		return (SubProcessTask[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getSubprocessTasks()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.Process spagicMetaDBProcess = AuditDBUtils.getProcessById(aSession, processId);
						if (spagicMetaDBProcess == null) {
							throw new SpagicWorkflowException(" Process with ID [" + processId + "] Not Found");
						}

						SubProcessTask[] result = getSubProcessTasks(spagicMetaDBProcess);
						return result;
					}
				},null); 
	}
	

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.spagic.workflow.api.IQueryAPI#getSubprocessTasks(java.lang.String)
	 */
	public SubProcessTask[] getSubprocessTasks(final String processName) {
		return (SubProcessTask[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getSubprocessTasks()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.Process spagicMetaDBProcess = ProcessEngine.getLastDefinitionForProcess(aSession,processName);
						if (spagicMetaDBProcess == null) {
							throw new SpagicWorkflowException(" Process with name [" + processName + "] Not Found");
						}

						SubProcessTask[] result = getSubProcessTasks(spagicMetaDBProcess);
						return result;
					}
				},null); 
	
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.spagic.workflow.api.IQueryAPI#getSubprocessTasks(org.spagic.workflow
	 * .api.Process)
	 */
	public SubProcessTask[] getSubprocessTasks(Process process) {
		return getSubprocessTasks(process.getProcessId());
	}

	/**
	 * Gets the sub process tasks.
	 * 
	 * @param spagicMetaDBProcess
	 *          the spagic meta db process
	 * 
	 * @return the sub process tasks
	 */
	public SubProcessTask[] getSubProcessTasks(org.spagic.metadb.model.Process spagicMetaDBProcess) {
		Set<SubProcess> subProcesses = spagicMetaDBProcess.getSubProcessesForIdProcess();

		List<SubProcessTask> toReturnTasks = new ArrayList<SubProcessTask>();

		Step step = null;
		org.spagic.metadb.model.Process theMetaDbSubProcess = null;
		SubProcessTask t = null;
		for (SubProcess subProcess : subProcesses) {
			step = subProcess.getStepByIdStep();
			theMetaDbSubProcess = subProcess.getProcessByIdSubprocess();
			t = new SubProcessTask();
			t.setTaskId(step.getIdStep());
			t.setTaskName(ProcessEngine.getShortTaskName(step.getName()));
			t.setSubProcessId(theMetaDbSubProcess.getIdProcess());
			t.setSubProcessName(theMetaDbSubProcess.getName());
			toReturnTasks.add(t);

		}
		SubProcessTask[] toReturnProcessesArray = new SubProcessTask[toReturnTasks.size()];
		return (SubProcessTask[]) toReturnTasks.toArray(toReturnProcessesArray);
	}

	/**
	 * Gets the names of the human tasks
	 * 
	 * @param spagicMetaDBProcess
	 *          the spagic meta db process
	 * 
	 * @return the sub process tasks
	 */
	public String[] getNextTasksNames(final long processInstanceId) {
		return (String[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getSubprocessTasks()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						List<TransitionInstance> l = AuditDBUtils.getTransitionInstancesActiveJbpm(aSession, processInstanceId);
						// getTransitionInstanceById(aSession, id);
						// (aSession,
						// spagicProcess.getIdProcess(),ProcessEngine.getStepLocalNamePart(processName,
						// spagicProcess.getProcessVersion(), taskName));
						String[] toReturnArray = new String[l.size()];
						int index = 0;
						for (TransitionInstance curTransitionIntance : l) {
							toReturnArray[index++] = ProcessEngine.getShortTaskName(curTransitionIntance.getStepByIdStepTarget().getName());
						}
						return toReturnArray;
					}
				},null); 
	}
	

	public DataHandler getTaskDocumentation(final String processName, final String taskName) {
		return (DataHandler)ProcessEngine.doQueryWithinSpagic("QueryAPI.getTaskDocumentation()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.Process spagicProcess = ProcessEngine.getLastDefinitionForProcess(aSession, processName);
							Step step = AuditDBUtils.getNamedStepForProcess(aSession, spagicProcess.getIdProcess(), ProcessEngine
									.getStepLocalNamePart(processName, spagicProcess.getProcessVersion(), taskName));
							ByteArrayDataSource bads = new ByteArrayDataSource(step.getDocumentation(), step.getDocumentationMimeType());
							bads.setName(step.getDocumentationFileName());
							DataHandler dataHandler = new DataHandler(bads);
							return dataHandler;
					}
				},null); 
	}
	

	public TaskInstance[] getHumanTaskInstances() {
		TaskInstance[] returnedTaskInstances = null;
		ArrayList<TaskInstance> tiList = new ArrayList<TaskInstance>();
		ProcessInstance[] processesInstances = getProcessInstances();
		for (int j = 0; j < processesInstances.length; j++) {
			long processInstanceId = processesInstances[j].getProcessInstanceId();
			TaskInstance[] taskInstances = getHumanTaskInstances(processInstanceId);
			for (int i = 0; i < taskInstances.length; i++) {
				tiList.add(taskInstances[i]);
			}
		}
		if (tiList.size() > 0)
			returnedTaskInstances = tiList.toArray(returnedTaskInstances);
		return returnedTaskInstances;
	}

	
	public TaskInstance[] getGeneratedHumanTaskInstances() {
		return getGeneratedHumanTaskInstances(null, null, null, null, null, null, 0, 0L);
	}

	public TaskInstance[] getGeneratedHumanTaskInstances(long processInstanceId) {
		return getGeneratedHumanTaskInstances(null, null, null, null, null, null, 0, processInstanceId);
	}

	public TaskInstance[] getGeneratedHumanTaskInstances(final String processName, 
														 final String taskName, 
														 final Date createFrom,
														 final Date createTo, 
														 final Date dueDateFrom, 
														 final Date dueDateTo, 
														 final int state, 
														 final long processInstanceId) {
		

		ProcessInstanceExecutionContext iec = null;
		if (processInstanceId > 0)
			iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		return (TaskInstance[]) ProcessEngine.doQueryWithinJBPMAndSpagic("QueryAPI.getGeneratedHumanTaskInstances()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
					
						Session jbpmSession = jbpmContext.getSession();
						Criteria criteria = jbpmSession.createCriteria(org.jbpm.taskmgmt.exe.TaskInstance.class);
						criteria.add(Restrictions.isNull("end"));

						completeCriteria(criteria, processName, taskName, createFrom, createTo, dueDateFrom, dueDateTo, state,
							processInstanceId, instanceExecutionContext);

						@SuppressWarnings("unchecked")
						List<org.jbpm.taskmgmt.exe.TaskInstance> jbpmTaskInstances = criteria.list();
						ArrayList<TaskInstance> tiList = new ArrayList<TaskInstance>();
						for (org.jbpm.taskmgmt.exe.TaskInstance jbpmTaskInstanceForTask : jbpmTaskInstances) {
							String processDefName = jbpmTaskInstanceForTask.getProcessInstance().getProcessDefinition().getName();
							Process process = getProcess(processDefName);
							org.spagic.metadb.model.ProcessInstance spagicPi = jbpmEngine.getSpagicProcessInstanceFromJBPMProcessInstance(jbpmTaskInstanceForTask.getProcessInstance(), spagicSession); 
							TaskInstance ti = jbpmEngine.getSpagicTaskInstanceFromJBPMProcessInstance(spagicSession,jbpmTaskInstanceForTask, process, spagicPi.getIdProcessInstance(), getTaskCompletionPercentage(jbpmTaskInstanceForTask));
							tiList.add(ti);
						
						}
						TaskInstance[] returnedTaskInstances = null;
						if (tiList.size() > 0)
							returnedTaskInstances = tiList.toArray(new TaskInstance[tiList.size()]);
						else
							// Saint - Wed Mar 10 16:52:25 2010
							// 
							// If the list is empty, create a 0 length array.
							returnedTaskInstances = new TaskInstance[0];
							//  </Saint>

						return returnedTaskInstances;
					}
				}, 
				iec);
		
			
		
	}
	
	

	
	private void completeCriteria(Criteria criteria, String processName, String taskName, Date createFrom, Date createTo,
			Date dueDateFrom, Date dueDateTo, int state, long processInstanceId, ProcessInstanceExecutionContext iec) {
		

		if (createFrom != null) {
			if (createTo != null)
				criteria.add(Restrictions.between("create", createFrom, createTo));
			else {
				criteria.add(Restrictions.ge("create", createFrom));
			}
		} else {
			if (createTo != null)
				criteria.add(Restrictions.le("create", createTo));
		}
		if (dueDateFrom != null) {
			if (dueDateTo != null)
				criteria.add(Restrictions.between("dueDate", dueDateFrom, dueDateTo));
			else {
				criteria.add(Restrictions.ge("dueDate", dueDateFrom));
			}
		} else {
			if (dueDateTo != null)
				criteria.add(Restrictions.le("dueDate", dueDateTo));
		}
		if (taskName != null)
			criteria.add(Restrictions.eq("name", taskName));
		if (processName != null)
			criteria.createCriteria("processInstance").createCriteria("processDefinition").add(
					Restrictions.eq("name", processName));
		if (state == 1)
			criteria.add(Restrictions.isNull("start"));
		if (state == 4)
			criteria.add(Restrictions.isNotNull("start"));
		if (processInstanceId > 0) {
			criteria.add(Restrictions.eq("processInstance.id", iec.getJbpmProcessInstanceId()));
		}

	}

	public Task getHumanTask(final long processId, final String taskName) {
		return (Task)ProcessEngine.doQueryWithinSpagic("QueryAPI.getHumanTask()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						Process aProcess = getProcess(aSession, processId);
						return getTask(aSession, aProcess, taskName);
					}
				},null); 
	}

	public Task getTask(Session aSession, Process aProcess, String taskName) {
		Task task = new Task();

		Step step = AuditDBUtils.getNamedStepForProcess(aSession, aProcess.getProcessId(), ProcessEngine
				.getStepLocalNamePart(aProcess.getName(), aProcess.getVersion(), taskName));

		task.setTaskId(step.getIdStep());
		// task.setInternalId(jbpmTask.getId());
		task.setTaskName(taskName);
		task.setProcessName(aProcess.getName());
		task.setProcessId(aProcess.getProcessId());

		// Retrieve the step property that contains the form information
		Property propertyFormId = AuditDBUtils.getStepPropertyByName(aSession, step, PROPERTY_FORM_ID);
		if (propertyFormId != null) {
			
			// Dynamic form task 
			Property propertyFormType = AuditDBUtils.getStepPropertyByName(aSession, step, PROPERTY_FORM_TYPE);
			Property propertyFormTitle = AuditDBUtils.getStepPropertyByName(aSession, step, PROPERTY_FORM_TITLE);
			
			// Orbeon FormRunner form task properties
			Property propertyFormApplication = AuditDBUtils.getStepPropertyByName(aSession, step, PROPERTY_FORM_APPLICATION);
			Property propertyFormForm = AuditDBUtils.getStepPropertyByName(aSession, step, PROPERTY_FORM_FORM);
			
			// Orbeon PDF form rendering properties
			Property propertyFormTask = AuditDBUtils.getStepPropertyByName(aSession, step, PROPERTY_FORM_TASK);
			
			if (propertyFormTitle != null) {
				String title =propertyFormTitle.getValue();
				if (title != null && title.length() > 0) {
					task.setFormTitle(title);
				}
			}
			
			if (propertyFormApplication != null) {
				String application = propertyFormApplication.getValue();
				if (application != null && application.length() >0) {
					task.setFormApplication(application);
				}
			}
			
			if (propertyFormForm != null) {
				String form = propertyFormForm.getValue();
				if (form != null && form.length() >0) {
					task.setFormForm(form);
				}
			}

			if (propertyFormTask != null) {
				String aTask =propertyFormTask.getValue();
				if (aTask != null && aTask.length() > 0) {
					task.setFormTask(aTask);
				}
			}

			// Fill form information
			String schemaIdentifier = propertyFormId.getValue();

			log.debug("schema metadata id = "+ schemaIdentifier);
			
			// When handling Orbeon forms we need no real metadata
			SchemaMetadata md=null;

			SchemaDefinition sd=null;
			if("___NoIdRequired__".equals(schemaIdentifier))
				sd = new SchemaDefinition();
			else  {
				try {
					md = AuditDBUtils.getSchemaDefinition(aSession, schemaIdentifier);
					if ( md == null ) {
						log.warn("Schema metadata is null for schema " + schemaIdentifier);
						sd = new SchemaDefinition();
					}
					else
						sd = SchemaDefinition.parseSchema(md.getDataSchema());
					
				} catch (Exception e) {
					throw new SpagicWorkflowException("Unable to retrieve schema definition with id: " + schemaIdentifier, e);
				}
			}	
			String formTypeString = (propertyFormType == null) ? "" : propertyFormType.getValue();
			Integer formTypeNumber = null;

			// Saint - Fri Dec 11 15:51:17 2009
			// 
			// Aggiunti gli else per non eseguire gli if non
			// richiesti. Aggiunto il PdfForm

			if (formTypeString.equals("NoForm")) 
				formTypeNumber = Task.NO_FORM;

			else if (formTypeString.equals("Dynamic"))
				formTypeNumber = Task.FORM_DYNAMIC;

			else if (formTypeString.equals("Flex"))
				formTypeNumber = Task.FORM_FLEX;

			else if (formTypeString.equals("OrbeonForm"))
				formTypeNumber = Task.FORM_ORBEON;

			else if (formTypeString.equals("PdfForm"))
				formTypeNumber = Task.FORM_PDF;
			//  </Saint>

			task.setFormType(formTypeNumber);
			if(sd != null)
			sd.setName(schemaIdentifier);
			task.setSchemaDefinition(sd);
		
		}

		return task;
	}

	/* (non-Javadoc)
	 * @see org.spagic.workflow.api.IQueryAPI#getProcessProperties(long)
	 */
	public Variable[] getProcessProperties(final long processId) {
		return (Variable[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcessProperties()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						List<Property> properties = AuditDBUtils.getProcessProperty(aSession, processId);
						Variable[] result = new Variable[properties.size()];
						int i = 0;
						for (Property prop : properties) {
							Variable var = new Variable(prop.getCode(), prop.getValue());
							result[i++] = var;				
						}			
						return result;
					
					}
				},null); 
	
	}

	/* (non-Javadoc)
	 * @see org.spagic.workflow.api.IQueryAPI#getProcessProperties(java.lang.String)
	 */
	public Variable[] getProcessProperties(final String processName) {
		
		return (Variable[])ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcessProperties()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.spagic.metadb.model.Process spagicMetaDBProcess = ProcessEngine.getLastDefinitionForProcess(spagicSession,
								processName);

						List<Property> properties = AuditDBUtils.getProcessProperty(spagicSession, spagicMetaDBProcess.getIdProcess());
						Variable[] result = new Variable[properties.size()];
						int i = 0;
						for (Property prop : properties) {
							Variable var = new Variable(prop.getCode(), prop.getValue());
							result[i++] = var;				
						}			
						return result;
					}
		},null);
					
	}
	
	
	
	public TaskInstance[] getHumanTaskInstancesAssignedToActor(final String actorId) {
		
		if (jbpmEngine.isSuperUser(actorId))
			return getGeneratedHumanTaskInstances(null, null, null, null, null, null,4, 0L);
		
			 return (TaskInstance[])ProcessEngine.doQueryWithinJBPMAndSpagic("QueryAPI.getHumanTaskInstancesAssignedToActor()", 
				 jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						Session jbpmSession = jbpmContext.getSession();
						Criteria criteria = jbpmSession.createCriteria(org.jbpm.taskmgmt.exe.TaskInstance.class);
						criteria.add(Restrictions.isNull("end"));

						criteria.add(Restrictions.eq("actorId", actorId));
						

						@SuppressWarnings("unchecked")
						List<org.jbpm.taskmgmt.exe.TaskInstance> jbpmTaskInstances = criteria.list();
						ArrayList<TaskInstance> tiList = new ArrayList<TaskInstance>();
						for (org.jbpm.taskmgmt.exe.TaskInstance jbpmTaskInstanceForTask : jbpmTaskInstances) {
							String processDefName = jbpmTaskInstanceForTask.getProcessInstance().getProcessDefinition().getName();
							Process process = getProcess(processDefName);
							org.spagic.metadb.model.ProcessInstance spagicPi = ProcessEngine.getSpagicProcessInstanceFromJBPMProcessInstance(jbpmTaskInstanceForTask.getProcessInstance(), aSession); 
							TaskInstance ti = jbpmEngine.getSpagicTaskInstanceFromJBPMProcessInstance(aSession,jbpmTaskInstanceForTask, process, spagicPi.getIdProcessInstance(), getTaskCompletionPercentage(jbpmTaskInstanceForTask));
							tiList.add(ti);
						
						}
						TaskInstance[] returnedTaskInstances = null;
					//	if (tiList.size() > 0)
							returnedTaskInstances = tiList.toArray(new TaskInstance[tiList.size()]);
						return returnedTaskInstances;
					}
		},null);
		
	
	}
		


}
